import os
import re
from pathlib import Path
from functools import partial
import xml.etree.ElementTree as ET
import math
import sys
import json 
import logging      
import types
import traceback

from ._utils.env import ARCGIS_ENABLE_TF_BACKEND

import_exception = None
try:
    import arcgis
    import numpy as np
    from fastai.vision.data import imagenet_stats, ImageList, bb_pad_collate
    from fastai.vision.transform import crop, rotate, dihedral_affine, brightness, contrast, skew, rand_zoom, get_transforms, flip_lr, ResizeMethod
    from fastai.vision import ImageDataBunch, parallel
    import fastai.vision
    from fastai.torch_core import data_collate
    import torch
    from .models._unet_utils import ArcGISSegmentationItemList, is_no_color
    from .models._maskrcnn_utils import ArcGISInstanceSegmentationItemList, ArcGISInstanceSegmentationMSItemList
    from ._utils._ner_utils import _NERData
    from ._utils.pascal_voc_rectangles import ObjectDetectionItemList
    from .models._superres_utils import resize_one
    from ._utils.common import ArcGISMSImage, ArcGISImageList
    from ._utils.env import HAS_GDAL, raise_gdal_import_error
    from ._utils.classified_tiles import show_batch_classified_tiles
    from ._utils.labeled_tiles import show_batch_labeled_tiles
    from ._utils.rcnn_masks import show_batch_rcnn_masks
    from ._utils.pascal_voc_rectangles import ObjectMSItemList, show_batch_pascal_voc_rectangles
    from ._utils.pointcloud_data import pointcloud_prepare_data
    from ._utils.superres import ImageImageListSR
    from fastai.tabular import TabularDataBunch
    from fastai.tabular.transform import FillMissing, Categorify, Normalize
    from fastai.tabular import cont_cat_split, add_datepart
    from ._utils.tabular_data import TabularDataObject
    from ._utils.text_data import TextDataObject
    from ._utils.cyclegan import ImageTupleList, prepare_data_ms_cyclegan
    from ._utils.pix2pix import ImageTupleList2, prepare_data_ms_pix2pix
    import random
    import PIL
    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False


band_abrevation_lib = {
    'b': 'BLUE',
    'c': 'CIRRUS',
    'ca': 'COASTAL AEROSOL',
    'g': 'GREEN',
    'nir': 'NEAR INFRARED',
    'nnir': 'NARROW NEAR INFRARED',
    'p': 'PANCHROMATIC',
    'r': 'RED',
    'swir': 'SHORT WAVELENGTH INFRARED',
    'swirc': 'SHORT WAVELENGTH INFRARED – Cirrus',
    'tir': 'THERMAL INFRARED',
    'vre': 'Vegetation red edge',
    'wv': 'WATER VAPOUR'
}

imagery_type_lib = {
    'landsat8': {
        "bands": ['ca', 'b', 'g', 'r', 'nir', 'swir', 'swir', 'c', 'qa', 'tir', 'tir'],
        "bands_info": { # incomplete
        }
    },
    "naip": {
        "bands": ['r', 'g', 'b', 'nir'],
        "bands_info": { # incomplete
        }
    },
    'sentinel2': { 
        "bands": ['ca', 'b', 'g', 'r', 'vre', 'vre', 'vre', 'nir', 'nnir', 'wv', 'swirc', 'swir', 'swir'],
        "bands_info": { # incomplete
            "b1": {
                "Name": "costal",
                "max": 10000,
                "min": 10000
            },            
            "b2": {
                "Name": "blue",
                "max": 10000,
                "min": 10000
            }
        }
    }
}


def get_installation_command():
    installation_steps = ("Install them using - 'conda install -c esri -c fastai -c pytorch arcgis=1.8.2 "
                          "scikit-image=0.15.0 pillow=6.2.2 libtiff=4.0.10 fastai=1.0.60 pytorch=1.4.0 "
                          "torchvision=0.5.0 scikit-learn=0.23.1 --no-pin'"
                          "\n'conda install gdal=2.3.3'"
                          "\n'pip install transformers==3.3.0'")
    return installation_steps 


def _raise_fastai_import_error(import_exception=import_exception):
    installation_steps = get_installation_command()
    raise Exception(f"""{import_exception} \n\nThis module requires fastai, PyTorch, torchvision and scikit-image as its dependencies.\n{installation_steps}""")


class _ImagenetCollater():
    def __init__(self, chip_size):
        self.chip_size = chip_size
    def __call__(self, batch):
        _xb = []
        for sample in batch:
            data = sample[0].data
            if data.shape[1] < self.chip_size or data.shape[2] < self.chip_size:
                data = sample[0].resize(self.chip_size).data
            _xb.append(data)
        _xb = torch.stack(_xb)
        _yb = torch.stack([torch.tensor(sample[1].data) for sample in batch])
        return _xb, _yb

def _bb_pad_collate(samples, pad_idx=0):
    "Function that collect `samples` of labelled bboxes and adds padding with `pad_idx`."
    if isinstance(samples[0][1], int):
        return data_collate(samples)
    max_len = max([len(s[1].data[1]) for s in samples])
    bboxes = torch.zeros(len(samples), max_len, 4)
    labels = torch.zeros(len(samples), max_len).long() + pad_idx
    imgs = []
    for i,s in enumerate(samples):
        imgs.append(s[0].data[None])
        bbs, lbls = s[1].data

        if not (bbs.nelement() == 0) or list(bbs) == [[0,0,0,0]]:
            bboxes[i,-len(lbls):] = bbs
            labels[i,-len(lbls):] = torch.tensor(lbls, device=bbs.device).long()
    return torch.cat(imgs,0), (bboxes,labels)    


def _get_bbox_classes(label_file, class_mapping , height_width=[], **kwargs):
    dataset_type = kwargs.get('dataset_type', None)

    if dataset_type == 'KITTI_rectangles':
        classes = []
        truncated = []
        occluded = []
        obs_angle = []
        bboxes = []
        hwl = []
        d_xyz = []
        occluded = []
        rot_yaxis = []     
        start_space = re.compile('^\s+') #pattern to capture leading spaces
        spaces_to_be_replaced = re.compile('(?<=\d)(\s+)(?=\d)') #pattern to capture spaces between numeric values

        with open(label_file) as f: #reading the bbox and class labels
            lines = f.readlines()
        
        for line in lines:
            line = re.sub(start_space, '', line)  
            lst = re.sub(spaces_to_be_replaced, ',', line).split(',') #reading kitti labels from string to a list
            xmin, ymin, xmax, ymax = [float(n) for n in lst[4:8]] #reading bbox coordinates
            hieght, width, length = [float(n) for n in lst[8:11]] 
            x, y, z = [float(n) for n in lst[11:14]]

            data_class_text = str(lst[0])
            if (not data_class_text.isnumeric() and not class_mapping.get(data_class_text))\
                or (data_class_text.isnumeric() and not(class_mapping.get(data_class_text) or class_mapping.get(int(data_class_text)))):
                continue
            if data_class_text.isnumeric():
                data_class_mapping = class_mapping[data_class_text] if class_mapping.get(data_class_text) else class_mapping[int(data_class_text)]
            else:
                data_class_mapping = class_mapping[data_class_text]
            
            classes.append(data_class_mapping) #object class
            truncated.append(float(lst[1]))    #if the object is truncated
            obs_angle.append(float(lst[2]))    #onservation angle
            occluded.append(float(lst[3]))     #if the object is occluded
            bboxes.append([ymin, xmin, ymax, xmax])
            height_width.append(((xmax - xmin)*1.25, (ymax - ymin)*1.25))
            hwl.append([hieght, width, length])
            d_xyz.append([x, y, z])
            rot_yaxis.append(float(lst[14]))   #angle of rotation along y axis
    else:
        tree = ET.parse(label_file)
        xmlroot = tree.getroot()
        bboxes = []
        classes = []
        for tag_obj in xmlroot.findall('object'):
            bnd_box = tag_obj.find('bndbox')
            xmin, ymin, xmax, ymax = float(bnd_box.find('xmin').text), \
                                    float(bnd_box.find('ymin').text), \
                                    float(bnd_box.find('xmax').text), \
                                    float(bnd_box.find('ymax').text)
            data_class_text = tag_obj.find('name').text

            if (not data_class_text.isnumeric() and not class_mapping.get(data_class_text))\
                or (data_class_text.isnumeric() and not(class_mapping.get(data_class_text) or class_mapping.get(int(data_class_text)))):
                continue

            if data_class_text.isnumeric():
                data_class_mapping = class_mapping[data_class_text] if class_mapping.get(data_class_text) else class_mapping[int(data_class_text)]
            else:
                data_class_mapping = class_mapping[data_class_text]

            classes.append(data_class_mapping)
            bboxes.append([ymin, xmin, ymax, xmax])
            height_width.append(((xmax - xmin)*1.25, (ymax - ymin)*1.25))

    if len(bboxes) == 0:
        return [[[0., 0., 0., 0.]], [list(class_mapping.values())[0]]]
    return [bboxes, classes]


def _get_bbox_lbls(imagefile, class_mapping, height_width, **kwargs):
    dataset_type = kwargs.get('dataset_type', None)
    if dataset_type == 'KITTI_rectangles':
        label_suffix = '.txt'
    else:
        label_suffix = '.xml'
    label_file = imagefile.parents[1] / 'labels' / imagefile.name.replace('{ims}'.format(ims=imagefile.suffix), label_suffix)
    return _get_bbox_classes(label_file, class_mapping, height_width, dataset_type=dataset_type)


def _get_lbls(imagefile, class_mapping):
    xmlfile = imagefile.parents[1] / 'labels' / imagefile.name.replace('{ims}'.format(ims=imagefile.suffix), '.xml')
    return _get_bbox_classes(xmlfile, class_mapping)[1][0]


def _get_multi_lbls(imagefile):
    """
    Function that returns class labels for an image for multilabel classification.
    input: imagefile (Path)
    returns: labels (List[str])
    """
    xmlfile = imagefile.parents[1] / 'labels' / imagefile.name.replace('{ims}'.format(ims=imagefile.suffix), '.xml')
    labels = ET.parse(xmlfile).getroot().find('object').find('name').text
    labels = labels.split(',')
    return labels


def _check_esri_files(path):
    if os.path.exists(path / 'esri_model_definition.emd') \
        and os.path.exists(path / 'map.txt') \
            and os.path.exists(path / 'esri_accumulated_stats.json'):
        return True

    return False


def _get_class_mapping(path, **kwargs):
    '''getting class mapping from labels, incase esri definition files and class mapping is not provided'''
    dataset_type = kwargs.get('dataset_type', None)
    class_mapping = {}
    if dataset_type == 'KITTI_rectangles':
        start_space = re.compile('^\s+') #pattern to capture leading spaces
        spaces_to_be_replaced = re.compile('(?<=\d)(\s+)(?=\d)') #pattern to capture spaces between numeric values

        for txtfile in os.listdir(path): 
            if not txtfile.endswith('.txt'):
                continue
            with open(os.path.join(path , txtfile)) as f:
                lines = f.readlines()
            for line in lines:
                line = re.sub(start_space, '', line)
                lst = re.sub(spaces_to_be_replaced, ',', line).split(',')
                class_mapping[str(lst[0])] = str(lst[0])
    else:
        for xmlfile in os.listdir(path):
            if not xmlfile.endswith('.xml'):
                continue
            tree = ET.parse(os.path.join(path, xmlfile))
            xmlroot = tree.getroot()
            for tag_obj in xmlroot.findall('object'):
                class_mapping[tag_obj.find('name').text] = tag_obj.find('name').text

    return class_mapping


def _get_batch_stats(image_list, 
                     norm_pct=1, 
                     _band_std_values=False, 
                     scaled_std=True, 
                     reshape=True):
    n_normalization_samples = round(len(image_list)*norm_pct)
    #n_normalization_samples = max(256, n_normalization_samples)
    random_indexes = np.random.randint(0, len(image_list), size=min(n_normalization_samples, len(image_list)))

    # Original Band Stats
    min_values_store = []
    max_values_store = []
    mean_values_store = []

    data_shape = image_list[0].data.shape
    n_bands = data_shape[0]
    feasible_chunk = round(512*4*400/(n_bands*data_shape[1])) # ~3gb footprint
    chunk = min(feasible_chunk, n_normalization_samples)
    i = 0
    n_c = image_list[0].data.shape[0]
    for i in range(0, n_normalization_samples, chunk):
        if reshape:
            x_tensor_chunk = torch.cat([x.data.view(n_c, -1) for x in image_list[random_indexes[i:i+chunk]]], dim=1)
            min_values = x_tensor_chunk.min(dim=1).values
            max_values = x_tensor_chunk.max(dim=1).values
            mean_values = x_tensor_chunk.mean(dim=1)
        else:
            """
            min_values = torch.zeros(n_bands)
            max_values = torch.zeros(n_bands)
            mean_values = torch.zeros(n_bands)
            for bi in range(n_bands):
                min_values[bi] = x_tensor_chunk[:, bi].min()
                max_values[bi] = x_tensor_chunk[:, bi].max()
                mean_values[bi] = x_tensor_chunk[:, bi].mean()
            """          
            x_tensor_chunk = torch.stack([x.data for x in image_list[random_indexes[i:i+chunk]]])
            min_values = x_tensor_chunk.min(dim=0)[0].min(dim=1)[0].min(dim=1)[0]
            max_values = x_tensor_chunk.max(dim=0)[0].max(dim=1)[0].max(dim=1)[0]
            mean_values = x_tensor_chunk.mean((0, 2, 3))
        min_values_store.append(min_values)
        max_values_store.append(max_values)
        mean_values_store.append(mean_values)

    band_max_values = torch.stack(max_values_store).max(dim=0)[0]
    band_min_values = torch.stack(min_values_store).min(dim=0)[0]
    band_mean_values = torch.stack(mean_values_store).mean(dim=0)
    
    view_shape = _get_view_shape(image_list[0].data, band_mean_values)

    if _band_std_values:
        std_values_store = []
        for i in range(0, n_normalization_samples, chunk):
            x_tensor_chunk = torch.stack([ x.data for x in image_list[random_indexes[i:i+chunk]] ] )
            std_values = (x_tensor_chunk - band_mean_values.view(view_shape)).pow(2).sum((0, 2, 3))
            std_values_store.append(std_values)
        band_std_values = (torch.stack(std_values_store).sum(dim=0) / ((n_normalization_samples * data_shape[1] * data_shape[2])-1)).sqrt()
    else:
        band_std_values = None

    # Scaled Stats
    scaled_min_values = torch.tensor([0 for i in range(n_bands)], dtype=torch.float32)
    scaled_max_values = torch.tensor([1 for i in range(n_bands)], dtype=torch.float32)
    scaled_mean_values = _tensor_scaler(band_mean_values, band_min_values, band_max_values, mode='minmax')
    
    if scaled_std:
        scaled_std_values_store = []
        for i in range(0, n_normalization_samples, chunk):
            x_tensor_chunk = torch.stack([ x.data for x in image_list[random_indexes[i:i+chunk]] ] )
            x_tensor_chunk = _tensor_scaler(x_tensor_chunk, band_min_values, band_max_values, mode='minmax')
            std_values = (x_tensor_chunk - scaled_mean_values.view(view_shape)).pow(2).sum((0, 2, 3))
            scaled_std_values_store.append(std_values)
        scaled_std_values = (torch.stack(scaled_std_values_store).sum(dim=0) / ((n_normalization_samples * data_shape[1] * data_shape[2])-1)).sqrt()
    else:
        scaled_std_values = None

    #return band_min_values, band_max_values, band_mean_values, band_std_values, scaled_min_values, scaled_max_values, scaled_mean_values, scaled_std_values
    return {
        "band_min_values": band_min_values, 
        "band_max_values": band_max_values, 
        "band_mean_values": band_mean_values, 
        "band_std_values": band_std_values, 
        "scaled_min_values": scaled_min_values, 
        "scaled_max_values": scaled_max_values, 
        "scaled_mean_values": scaled_mean_values, 
        "scaled_std_values": scaled_std_values
    }

def sniff_rgb_bands(band_names):
    band_mapping_reverse = {k.lower(): i for i, k in enumerate(band_names)}
    rgb_bands = []
    for b in [
        'red',
        'green',
        'blue'
    ]:
        bi = band_mapping_reverse.get(b, None)
        if bi is None:
            return
        rgb_bands.append(bi)
    return rgb_bands

def data_is_multispectral(emd_info: dict) -> bool:
    """
    :param emd_info: Dictionary containing EMD info
    :return: Boolean value denoting whether data is multispectral or not.
    """
    is_multispectral = False
    if not 'InputRastersProps' in emd_info:
        return is_multispectral
    if len(emd_info['InputRastersProps']["BandNames"]) !=3:
        is_multispectral = True
    return is_multispectral

def _get_view_shape(tensor_batch, band_factors):
    view_shape = [1 for i in range(len(tensor_batch.shape))]
    view_shape[tensor_batch.shape.index(band_factors.shape[0])] = band_factors.shape[0]
    return tuple(view_shape)


def _tensor_scaler(tensor_batch, min_values, max_values, mode='minmax', create_view=True):
    if create_view:
        view_shape = _get_view_shape(tensor_batch, min_values)
        max_values = max_values.view(view_shape)
        min_values = min_values.view(view_shape)
    if mode=='minmax':
        # new_value = (((old_value - old_min) / (old_max-old_min))*(new_max-new_min))+new_min
        scaled_tensor_batch = (tensor_batch - min_values) / ( (max_values - min_values) + 1e-05)
    return scaled_tensor_batch


def _tensor_scaler_tfm(tensor_batch, min_values, max_values, mode='minmax'):
    x = tensor_batch[0]
    y = tensor_batch[1]
    max_values = max_values.view(1, -1, 1, 1).to(x.device)
    min_values = min_values.view(1, -1, 1, 1).to(x.device)
    x = _tensor_scaler(x, min_values, max_values, mode, create_view=False)
    return (x, y)


def _extract_bands_tfm(tensor_batch, band_indices):
    x_batch = tensor_batch[0][:, band_indices]
    y_batch = tensor_batch[1]
    return (x_batch, y_batch)


def prepare_textdata(
        path,
        task,
        text_columns,
        label_columns,
        train_file="train.csv",
        valid_file=None,
        val_split_pct=0.1,
        seed=42,
        batch_size=8,
        process_labels=False,
        remove_html_tags=False,
        remove_urls=False
    ):
    """
    Prepares a text data object from the files present at data folder

    =====================   =================================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------------
    path                    Required directory path. The directory path where
                            the training and validation files are present.
    ---------------------   -------------------------------------------------
    task                    Required string. The task for which the dataset is
                            prepared. Available choice at this point is "classification"
                            and "sequence_translation".
    ---------------------   -------------------------------------------------
    text_columns            Required string. The column that will be used as
                            feature.
    ---------------------   -------------------------------------------------
    label_columns           Required list. The list of columns denoting the
                            class label/translated text to predict. Provide a list of columns
                            in case of multi-label classification problem
    ---------------------   -------------------------------------------------
    train_file              Optional string. The file name containing the
                            training data. Supported file formats/extensions are
                            .csv and .tsv
                            Default value is `train.csv`
    ---------------------   -------------------------------------------------
    valid_file              Optional string. The file name containing the
                            validation data. Supported file formats/extensions
                            are .csv and .tsv.
                            Default value is `None`. If None then some portion
                            of the training data will be kept for validation
                            (based on the value of `val_split_pct` parameter)
    ---------------------   -------------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
                            By default 10% data is kept for validation.
    ---------------------   -------------------------------------------------
    seed                    Optional integer. Random seed for reproducible
                            train-validation split.
                            Default value is 42.
    ---------------------   -------------------------------------------------
    batch_size              Optional integer. Batch size for mini batch gradient
                            descent (Reduce it if getting CUDA Out of Memory
                            Errors).
                            Default value is 16.
    ---------------------   -------------------------------------------------
    process_labels          Optional boolean. If true, default processing functions
                            will be called on label columns as well.
                            Default value is False.
    ---------------------   -------------------------------------------------
    remove_html_tags        Optional boolean. If true, remove html tags from text.
                            Default value is False.
    ---------------------   -------------------------------------------------
    remove_urls             Optional boolean. If true, remove urls from text.
                            Default value is False.
    =====================   =================================================

    :returns: `TextData` object

    """
    # allowed_tasks = ["classification", "summarization", "translation",
    #                  "question-answering", "ner", "text-generation"]

    if not HAS_FASTAI:
        _raise_fastai_import_error(import_exception)

    # if task not in allowed_tasks:
    #     raise Exception(f"Wrong task choosen. Allowed tasks are {allowed_tasks}")

    if isinstance(label_columns, (str, bytes)):
        label_columns = [label_columns]

    force_cpu = arcgis.learn.models._arcgis_model._device_check()

    if hasattr(arcgis, "env") and force_cpu == 1:
        arcgis.env._processorType = "CPU"

    if task == "classification":
        return TextDataObject.prepare_data_for_classification(
            path,
            text_columns,
            label_columns,
            train_file=train_file,
            valid_file=valid_file,
            val_split_pct=val_split_pct,
            seed=seed,
            batch_size=batch_size,
            process_labels=process_labels,
            remove_html_tags=remove_html_tags,
            remove_urls=remove_urls
        )
    
    elif task.lower() == "sequence_translation":
        return TextDataObject.prepare_data_for_seq2seq(
            path,
            text_columns,
            label_columns,
            train_file=train_file,
            val_split_pct=val_split_pct,
            seed=seed,
            batch_size=batch_size,
            process_labels=process_labels,
            remove_html_tags=remove_html_tags,
            remove_urls=remove_urls
        )

    else:
        logger = logging.getLogger()
        logger.error(f"Wrong task - {task} provided. This function can handle only `classification` and 'sequence_translation' task currently")
        raise Exception(f"Wrong task - {task} provided. This function can handle only `classification` and 'sequence_translation' task currently")


def prepare_tabulardata(
        input_features=None,
        variable_predict=None,
        explanatory_variables=None,
        explanatory_rasters=None,
        date_field=None,
        distance_features=None,
        preprocessors=None,
        val_split_pct=0.1,
        seed=42,
        batch_size=64,
        index_field=None
    ):

    """
    Prepares a tabular data object from input_features and optionally rasters.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    input_features          Optional Feature Layer Object or spatially enabled dataframe.
                            This contains features denoting the value of the dependent variable.
                            Leave empty for using rasters with MLModel.
    ---------------------   -------------------------------------------
    variable_predict        Optional String, denoting the field_name of
                            the variable to predict.
                            Keep none for unsupervised training using MLModel.
    ---------------------   -------------------------------------------
    explanatory_variables   Optional list containing field names from input_features
                            By default the field type is continuous.
                            To override field type to categorical, pass
                            a 2-sized tuple in the list containing:
                                1. field to be taken as input from the input_features.
                                2. True/False denoting Categorical/Continuous variable.
                            For example:
                                ["Field_1", ("Field_2", True)]
                            Here Field_1 is treated as continuous and
                            Field_2 as categorical.
    ---------------------   -------------------------------------------
    explanatory_rasters     Optional list containing Raster objects.
                            By default the rasters are continuous.
                            To mark a raster categorical, pass a 2-sized tuple containing:
                                1. Raster object.
                                2. True/False denoting Categorical/Continuous variable.
                            For example:
                                [raster_1, (raster_2, True)]
                            Here raster_1 is treated as continuous and
                            raster_2 as categorical.
    ---------------------   -------------------------------------------
    date_field              Optional field_name.
                            This field contains the date in the input_features.
                            The field type can be a string or date time field.
                            If specified, the field will be split into
                            Year, month, week, day, dayofweek, dayofyear,
                            is_month_end, is_month_start, is_quarter_end,
                            is_quarter_start, is_year_end, is_year_start,
                            hour, minute, second, elapsed and these will be added
                            to the prepared data as columns.
                            All fields other than elapsed and dayofyear are treated
                            as categorical.
    ---------------------   -------------------------------------------
    distance_features       Optional list of Feature Layer objects.
                            Distance is calculated from features in these layers
                            to features in input_features.
                            Nearest distance to each feature is added in the prepared
                            data.
                            Field names in the prepared data added are
                            "NEAR_DIST_1", "NEAR_DIST_2" etc.
    ---------------------   -------------------------------------------
    preprocessors           For Fastai: Optional transforms list.
                            For Scikit-learn:
                            1. Supply a column transformer object.
                            2. Supply a list of tuple,
                            For example:
                            [('Col_1', 'Col_2', Transform1()), ('Col_3', Transform2())]
                            Categorical data is by default encoded.
                            If nothing is specified, default transforms are applied
                            to fill missing values and normalize categorical data.
                            For Raster use raster.name for the the first band,
                            raster.name_1 for 2nd band, raster.name_2 for 3rd
                            and so on.
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
                            By default 10% data is kept for validation.
    ---------------------   -------------------------------------------
    seed                    Optional integer. Random seed for reproducible
                            train-validation split.
                            Default value is 42.
    ---------------------   -------------------------------------------
    batch_size              Optional integer. Batch size for mini batch gradient
                            descent (Reduce it if getting CUDA Out of Memory
                            Errors).
                            Default value is 64.
    ---------------------   -------------------------------------------
    index_field             Optional string. Field Name in the input features
                            which will be used as index field for the data.
                            Used for Time Series, to visualize values on the
                            x-axis.
    =====================   ===========================================

    :returns: `TabularData` object

    """
    if input_features is None and (explanatory_rasters is None or len(explanatory_rasters) == 0):
        raise Exception("No Features or Rasters found")

    import warnings
    if not HAS_FASTAI:
        _raise_fastai_import_error(import_exception)

    force_cpu = arcgis.learn.models._arcgis_model._device_check()

    if hasattr(arcgis, "env") and force_cpu == 1:
        arcgis.env._processorType = "CPU"

    HAS_COLUMN_TRANSFORMS = False

    column_transforms_mapping = {}
    if preprocessors and isinstance(preprocessors, list):
        for transform in preprocessors:
            if isinstance(transform, tuple):
                HAS_COLUMN_TRANSFORMS = True
                break

        if HAS_COLUMN_TRANSFORMS:
            column_transforms = []
            for transform in preprocessors:
                if not isinstance(transform, tuple):
                    warnings.warn("Please pass (Field_Name, transform) in the list of preprocessors")
                    return
                column_transforms.append(
                    (
                        transform[-1],
                        list(transform[0:-1])
                    )
                )

            from sklearn.compose import make_column_transformer
            preprocessors = make_column_transformer(*column_transforms)

    if preprocessors:
        for transform in preprocessors.transformers:
            for column in transform[2]:
                if not column_transforms_mapping.get(column):
                    column_transforms_mapping[column] = []
                if 'pipeline' in transform[0]:
                    for step in transform[1].steps:
                        column_transforms_mapping[column].append(step[1])
                else:
                    column_transforms_mapping[column].append(transform[1])

    return TabularDataObject.prepare_data_for_layer_learner(
        input_features,
        variable_predict,
        feature_variables=explanatory_variables,
        raster_variables=explanatory_rasters,
        date_field=date_field,
        distance_feature_layers=distance_features,
        procs=preprocessors,
        val_split_pct=val_split_pct,
        seed=seed,
        batch_size=batch_size,
        index_field=index_field,
        column_transforms_mapping=column_transforms_mapping
    )


def prepare_data(path,
                 class_mapping=None, 
                 chip_size=224, 
                 val_split_pct=0.1, 
                 batch_size=64, 
                 transforms=None, 
                 collate_fn=_bb_pad_collate, 
                 seed=42, 
                 dataset_type=None, 
                 resize_to=None,
                 **kwargs):
    """
    Prepares a data object from training sample exported by the 
    Export Training Data tool in ArcGIS Pro or Image Server, or training 
    samples in the supported dataset formats. This data object consists of 
    training and validation data sets with the specified transformations, 
    chip size, batch size, split percentage, etc. 
    -For object detection, use Pascal_VOC_rectangles or KITTI_rectangles format.
    -For feature categorization use Labelled Tiles or ImageNet format.
    -For pixel classification, use Classified Tiles format.
    -For entity extraction from text, use IOB, BILUO or ner_json formats. 

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    path                    Required string. Path to data directory.
    ---------------------   -------------------------------------------
    class_mapping           Optional dictionary. Mapping from id to
                            its string label.
                            For dataset_type=IOB, BILUO or ner_json:
                                Provide address field as class mapping
                                in below format:
                                class_mapping={'address_tag':'address_field'}.
                                Field defined as 'address_tag' will be treated
                                as a location. In cases where trained model extracts
                                multiple locations from a single document, that 
                                document will be replicated for each location.

    ---------------------   -------------------------------------------
    chip_size               Optional integer, default 224. Size of the image to train the
                            model. Images are cropped to the specified chip_size. If image size is less
                            than chip_size, the image size is used as chip_size. Not supported for superres.
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
    ---------------------   -------------------------------------------
    batch_size              Optional integer. Batch size for mini batch gradient
                            descent (Reduce it if getting CUDA Out of Memory
                            Errors). Batch size is required to be greater than 1.
    ---------------------   -------------------------------------------
    transforms              Optional tuple. Fast.ai transforms for data
                            augmentation of training and validation datasets
                            respectively (We have set good defaults which work
                            for satellite imagery well). If transforms is set
                            to `False` no transformation will take place and 
                            `chip_size` parameter will also not take effect.
                            If the dataset_type is 'PointCloud', use 
                            `Transform3d` class from `arcgis.learn`.
    ---------------------   -------------------------------------------
    collate_fn              Optional function. Passed to PyTorch to collate data
                            into batches(usually default works).
    ---------------------   -------------------------------------------
    seed                    Optional integer. Random seed for reproducible
                            train-validation split.
    ---------------------   -------------------------------------------
    dataset_type            Optional string. `prepare_data` function will infer 
                            the `dataset_type` on its own if it contains a 
                            map.txt file. If the path does not contain the 
                            map.txt file pass either of 'PASCAL_VOC_rectangles', 
                            'KITTI_rectangles', 'RCNN_Masks', 'Classified_Tiles', 
                            'Labeled_Tiles', 'MultiLabeled_Tiles', 'Imagenet',  
                            'PointCloud', 'ImageCaptioning', 'ChangeDetection',
                            'superres', 'CycleGAN' and 'Pix2Pix'.
                            This parameter is mandatory for data which are not
                            exported by ArcGIS Pro / Enterprise which includes
                            'PointCloud', 'ImageCaptioning', 'ChangeDetection',
                            'CycleGAN' and 'Pix2Pix'.
                            This parameter is also mandatory while preparing data
                            for 'EntityRecognizer' model. Accepted data format
                            for this model are - ['ner_json','BIO', 'LBIOU'].
    ---------------------   -------------------------------------------
    resize_to               Optional integer. Resize the images to a given size.
                            Works only for "PASCAL_VOC_rectangles" and "superres".
                            First resizes the image to the given size and
                            then crops images of size equal to chip_size.
                            Note: Keep chip_size < resize_to
    =====================   ===========================================

    **Keyword Arguments**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    imagery_type            Optional string. Type of imagery used to export 
                            the training data, valid values are:
                                - 'naip'
                                - 'sentinel2'
                                - 'landsat8'
                                - 'ms' - any other type of imagery
    ---------------------   -------------------------------------------
    bands                   Optional list. Bands of the imagery used to export 
                            training data. 
                            For example ['r', 'g', 'b', 'nir', 'u'] 
                            where 'nir' is near infrared band and 'u' is a miscellaneous band.
    ---------------------   -------------------------------------------
    rgb_bands               Optional list. Indices of red, green and blue bands 
                            in the imagery used to export the training data.
                            for example: [2, 1, 0] 
    ---------------------   -------------------------------------------
    extract_bands           Optional list. Indices of bands to be used for
                            training the model, same as in the imagery used to 
                            export the training data.
                            for example: [3, 1, 0] where we will not be using 
                            the band at index 2 to train our model. 
    ---------------------   -------------------------------------------
    norm_pct                Optional float. Percentage of training data to be 
                            used for calculating imagery statistics for  
                            normalizing the data. 
                            Default is 0.3 (30%) of data.
    ---------------------   -------------------------------------------
    downsample_factor       Optional integer. Factor to downsample the images 
                            for image SuperResolution. 
                            for example: if value is 2 and image size 256x256,
                            it will create label images of size 128x128.
                            Default is 4
    ---------------------   -------------------------------------------
    encoding                Optional string.
                            Applicable only when dataset_type=IOB, BILUO or ner_json:
                            The encoding to read the csv/json file.
                            Default is 'UTF-8'
    =====================   ===========================================

    :returns: data object

    """
    emd = {}
    height_width = []
    not_label_count = [0]

    if not HAS_FASTAI:
        _raise_fastai_import_error()

    (fastai.vision.data.image_extensions).add('.mrf')

    if isinstance(path, str) and not os.path.exists(path):
        message = f"Invalid input path. \nCould not find the path specified \n'{path}' \nPlease ensure that the input path is correct."
        if '\\' in path:
            message+=f"""\n\nif you are using windows style paths please ensure you have specified paths with raw modifier. for example {"path=r'{path}'"}"""
        raise Exception(message)

    if type(path) is str:
        path = Path(path)

    databunch_kwargs = {'num_workers':0} if sys.platform == 'win32' else {}
    databunch_kwargs['bs'] = batch_size

    force_cpu = arcgis.learn.models._arcgis_model._device_check()

    if hasattr(arcgis, "env") and force_cpu == 1:
        arcgis.env._processorType = "CPU"

    if getattr(arcgis.env, "_processorType", "") == "CPU":
        databunch_kwargs["device"] = torch.device('cpu')

    if ARCGIS_ENABLE_TF_BACKEND:
        databunch_kwargs["device"] = torch.device('cpu')
        databunch_kwargs["pin_memory"] = False

    kwargs_transforms = {}
    if resize_to:
        kwargs_transforms['size'] = resize_to
        # Applying SQUISH ResizeMethod to avoid reflection padding
        kwargs_transforms['resize_method'] = ResizeMethod.SQUISH

    has_esri_files = _check_esri_files(path)

    # For change detection export data using export tiles format.
    if dataset_type == 'ChangeDetection':
        from ._utils.change_detection_data import folder_check
        folder_check(path)
        json_file = path / 'images_before' / 'esri_model_definition.emd'
        if json_file.exists():
            with open(json_file) as f:
                emd = json.load(f)
        else:
            from ._utils.change_detection_data import get_files, image_extensions
            files_list = get_files(path  / 'images_before',
                                  extensions=image_extensions,
                                  recurse=True)
            msimage_list = ArcGISImageList(files_list)
            if msimage_list[0].shape[0] != 3:
                kwargs['imagery_type'] = 'ms'
    elif dataset_type == "CycleGAN" or dataset_type == "Pix2Pix":
        from ._utils.cyclegan import get_files, image_extensions
        path_a = path/"Images"/"train_a"
        path_b = path/"Images"/"train_b"
        files_list_a = get_files(path_a, extensions=image_extensions, recurse=True)
        files_list_b = get_files(path_b, extensions=image_extensions, recurse=True)
        msimage_list_a = ArcGISImageList(files_list_a)
        msimage_list_b = ArcGISImageList(files_list_b)
        if msimage_list_a[0].shape[0] != 3 or msimage_list_b[0].shape[0] != 3:
            kwargs['imagery_type'] = 'ms'
    alter_class_mapping = False
    color_mapping = None

    # Multispectral Kwargs init
    _bands = None
    _imagery_type = None
    _is_multispectral = False
    _show_batch_multispectral = None
    stats_file = path / 'esri_accumulated_stats.json'

    if dataset_type is None and not has_esri_files:
        raise Exception("Could not infer dataset type. Please specify a supported dataset type or ensure that the path contains valid esri files")
    elif dataset_type is None and has_esri_files:
        with open(stats_file) as f:
            stats = json.load(f)
            dataset_type = stats['MetaDataMode']
    
    if dataset_type not in ["Imagenet", "superres", "Export_Tiles"] and has_esri_files:
        with open(stats_file) as f:
            stats = json.load(f)
            dataset_type = stats['MetaDataMode']

        with open(path / 'map.txt') as f:
            while True:
                line = f.readline()
                if len(line.split())==2:
                    break
        try:
            img_size = ArcGISMSImage.open_gdal((path/(line.split()[0]).replace('\\', os.sep))).shape[-1]
        except:            
            img_size = PIL.Image.open((path/(line.split()[0]).replace('\\', os.sep))).size[-1]
        if chip_size > img_size:
            chip_size = img_size
        right = line.split()[1].split('.')[-1].lower()

        json_file = path / 'esri_model_definition.emd'
        with open(json_file) as f:
            emd = json.load(f)

        # Create Class Mapping from EMD if not specified by user
        ## Validate user defined class_mapping keys with emd (issue #3064)
        # Get classmapping from emd file.
        try:
            emd_class_mapping = {i['Value']: i['Name'] for i in emd['Classes']}
        except KeyError:
            emd_class_mapping = {i['ClassValue']: i['ClassName'] for i in emd['Classes']}

        ## Change all keys to int.
        if class_mapping is not None:
            class_mapping = {int(key):value for key, value in class_mapping.items()}
        else:
            class_mapping = {}

        ## Map values from user defined classmapping to emd classmapping.
        for key, _ in emd_class_mapping.items():
            if class_mapping.get(key) is not None:
                emd_class_mapping[key] = class_mapping[key]
            
        class_mapping = emd_class_mapping

        color_mapping = {(i.get('Value', 0) or i.get('ClassValue', 0)): i['Color'] for i in emd.get('Classes', [])}

        if color_mapping.get(None):
            del color_mapping[None]

        if class_mapping.get(None):
            del class_mapping[None]

        # Multispectral support from EMD 
        # Not Implemented Yet
        if emd.get('bands', None) is not None: 
            _bands = emd.get['bands'] # Not Implemented
            
        if emd.get('imagery_type', None) is not None: 
            _imagery_type = emd.get['imagery_type'] # Not Implemented        

    elif dataset_type in ['PASCAL_VOC_rectangles', 'KITTI_rectangles'] and not has_esri_files:
        if class_mapping is None:
            class_mapping = _get_class_mapping(path / 'labels', dataset_type=dataset_type)
            alter_class_mapping = True

    _map_space = "MAP_SPACE"
    _pixel_space = "PIXEL_SPACE"
    if has_esri_files:
        _image_space_used = emd.get('ImageSpaceUsed', _map_space)
    else:
        _image_space_used = _pixel_space

    # Multispectral check
    # With Python API for ArcGIS 1.9 multispectral workflow will automatically kick in with the following conditions
    # 1. If the imagery source is not having exactly three bands
    # 2. If there is any band other than RGB
    # 3. If None among all three bands in the imagery is unknown
    #
    imagery_type = 'ASSUMED_RGB'
    if kwargs.get('imagery_type', None) is not None:
        imagery_type = kwargs.get('imagery_type')
    elif _imagery_type is not None:
        imagery_type = _imagery_type
    if "InputRastersProps" in emd and kwargs.get('imagery_type', None) is None:
        sensor_name = emd["InputRastersProps"]["SensorName"]
        if len(emd["AllTilesStats"])!=3:
            if not (len(emd["AllTilesStats"]) == 4 and emd["InputRastersProps"]["BandNames"][3].lower() == 'alpha'):
                imagery_type = sensor_name
        else:
            # Check by band names
            band_mapping = {i:b.lower() for i, b in enumerate(emd["InputRastersProps"]["BandNames"])}
            for b in emd["WellKnownBandNames (FYI, these band names can be used in ExtractBands)"]:
                if b.lower() in [
                    "red",
                    "green",
                    "blue"
                ]:
                    continue
                if b.lower() in band_mapping:
                    imagery_type = sensor_name
                    break
            # Check by values
            for stat in emd["AllTilesStats"]:
                if stat["Min"] < 0 or stat["Max"] > 255:
                    imagery_type = sensor_name
                    break

    if (not imagery_type in ('ASSUMED_RGB', 'RGB')) and not HAS_GDAL:
        raise_gdal_import_error()

    bands = None
    if kwargs.get('bands', None) is not None:
        bands = kwargs.get('bands')
        for i, b in enumerate(bands):
            if type(b) == str:
                bands[i] = b.lower()
    elif imagery_type_lib.get(imagery_type, None) is not None:
        bands = imagery_type_lib.get(imagery_type)['bands']
    elif _bands is not None:
        bands = _bands

    rgb_bands = None
    if kwargs.get('rgb_bands', None) is not None:
        rgb_bands = kwargs.get('rgb_bands')
    elif bands is not None:
        rgb_bands = [ bands.index(b) for b in ['r', 'g', 'b'] if b in bands ]
    
    if (bands is not None) or (rgb_bands is not None) or (not imagery_type in ['RGB', 'ASSUMED_RGB']):
        if imagery_type in ['RGB', 'ASSUMED_RGB']:
            imagery_type = 'MULTISPECTRAL'
        _is_multispectral = True
    
    if kwargs.get('norm_pct', None) is not None:
        norm_pct= kwargs.get('norm_pct')
        norm_pct = min(max(0, norm_pct), 1)
    else:
        norm_pct = .3

    lighting_transforms = kwargs.get('lighting_transforms', True)

    if dataset_type == 'RCNN_Masks':

        def get_labels(x, label_dirs, ext=right):
            label_path = []
            for lbl in label_dirs:
                if os.path.exists(Path(lbl) / (x.stem + '.{}'.format(ext))):
                    label_path.append(Path(lbl) / (x.stem + '.{}'.format(ext)))
            return label_path

        label_dirs = []
        index_dir = {} #for handling calss value with any number
        for i, k in enumerate(sorted(class_mapping.keys())):
            label_dirs.append(class_mapping[k])
            index_dir[k] = i+1
        label_dir = [os.path.join(path/'labels', lbl) for lbl in label_dirs if os.path.isdir(os.path.join(path/'labels', lbl))]
        get_y_func = partial(get_labels, label_dirs= label_dir)

        def image_without_label(imagefile, not_label_count=[0]):
            label_mask = get_y_func(imagefile)
            if label_mask == []:
                not_label_count[0] += 1
                return False
            return True
        
        remove_image_without_label = partial(image_without_label, not_label_count=not_label_count)

        if class_mapping.get(0):
            del class_mapping[0]

        if color_mapping.get(0):
            del color_mapping[0]

        data = (ArcGISInstanceSegmentationItemList.from_folder(path/'images')
            .filter_by_func(remove_image_without_label)
            .split_by_rand_pct(val_split_pct, seed=seed)
            .label_from_func(get_y_func, chip_size=chip_size, classes=['NoData'] + list(class_mapping.values()), class_mapping=class_mapping, color_mapping=color_mapping, index_dir=index_dir))
        _show_batch_multispectral = show_batch_rcnn_masks
        
        if transforms is None:
            ranges = (0, 1)
            if _image_space_used == _map_space:
                train_tfms = [
                    crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                    dihedral_affine(),
                    brightness(change=(0.4, 0.6)),
                    contrast(scale=(1.0, 1.5)),
                    rand_zoom(scale=(1.0, 1.2))
                ]
            else:
                train_tfms = [
                    crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                    brightness(change=(0.4, 0.6)),
                    contrast(scale=(1.0, 1.5)),
                    rand_zoom(scale=(1.0, 1.2))
                ]
            val_tfms = [crop(size=chip_size, p=1., row_pct=0.5, col_pct=0.5)]
            transforms = (train_tfms, val_tfms)
            kwargs_transforms['size'] = chip_size

        kwargs_transforms['tfm_y'] = True

    elif dataset_type == 'Classified_Tiles':

        def get_y_func(x, ext=right):
            return x.parents[1] / 'labels' / (x.stem + '.{}'.format(ext))

        def image_without_label(imagefile, not_label_count=[0], ext=right):
            xmlfile = imagefile.parents[1] / 'labels' / (imagefile.stem + '.{}'.format(ext))
            if not os.path.exists(xmlfile):
                not_label_count[0] += 1
                return False
            return True
        
        remove_image_without_label = partial(image_without_label, not_label_count=not_label_count)

        if class_mapping.get(0):
            del class_mapping[0]

        if color_mapping.get(0):
            del color_mapping[0]

        if is_no_color(color_mapping):
            color_mapping = {j:[random.choice(range(256)) for i in range(3)] for j in class_mapping.keys()}
            
        # TODO : Handle NoData case

        data = ArcGISSegmentationItemList.from_folder(path/'images')\
            .filter_by_func(remove_image_without_label)\
            .split_by_rand_pct(val_split_pct, seed=seed)\
            .label_from_func(
                get_y_func, classes=(['NoData'] + list(class_mapping.values())),
                class_mapping=class_mapping,
                color_mapping=color_mapping
            )
        _show_batch_multispectral = show_batch_classified_tiles

        def classified_tiles_collate_fn(samples): # The default fastai collate_fn was causing memory leak on tensors
            r = ( torch.stack([x[0].data for x in samples]), torch.stack([x[1].data for x in samples]) )
            return r
        databunch_kwargs['collate_fn'] = classified_tiles_collate_fn

        if transforms is None:
            if _image_space_used == _map_space:
                transforms = get_transforms(
                    flip_vert=True,
                    max_rotate=90.,
                    max_zoom=3.0,
                    max_lighting=0.5
                )
            else:
                transforms = get_transforms(
                    max_zoom=3.0,
                    max_lighting=0.5
                )

        kwargs_transforms['tfm_y'] = True
        kwargs_transforms['size'] = chip_size
    elif dataset_type in ['PASCAL_VOC_rectangles', 'KITTI_rectangles']:

        def image_without_label(imagefile, dataset_type, not_label_count=[0]):
            if dataset_type == 'KITTI_rectangles':
                label_suffix='.txt'
            else:
                label_suffix='.xml'
            label_file = imagefile.parents[1] / 'labels' / imagefile.name.replace('{ims}'.format(ims=imagefile.suffix), label_suffix)
            if not os.path.exists(label_file):
                not_label_count[0] += 1
                return False
            return True
        remove_image_without_label = partial(image_without_label, not_label_count=not_label_count, dataset_type=dataset_type)
        get_y_func = partial(
            _get_bbox_lbls,
            class_mapping=class_mapping,
            height_width=height_width,
            dataset_type=dataset_type
        )

        data = ObjectDetectionItemList.from_folder(path / 'images') \
            .filter_by_func(remove_image_without_label) \
            .split_by_rand_pct(val_split_pct, seed=seed) \
            .label_from_func(get_y_func)
        _show_batch_multispectral = show_batch_pascal_voc_rectangles


        if transforms is None:
            ranges = (0, 1)
            if _image_space_used == _map_space:
                train_tfms = [
                    crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                    dihedral_affine(),
                    brightness(change=(0.4, 0.6)),
                    contrast(scale=(0.75, 1.5)),
                    rand_zoom(scale=(1.0, 1.5))
                ]
            else:
                train_tfms = [
                    crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                    brightness(change=(0.4, 0.6)),
                    contrast(scale=(0.75, 1.5)),
                    rand_zoom(scale=(1.0, 1.5))
                ]
            val_tfms = [crop(size=chip_size, p=1., row_pct=0.5, col_pct=0.5)]
            transforms = (train_tfms, val_tfms)

        kwargs_transforms['tfm_y'] = True
        databunch_kwargs['collate_fn'] = collate_fn
    elif dataset_type in ['Labeled_Tiles', 'MultiLabeled_Tiles', 'Imagenet']:
        if dataset_type == 'Labeled_Tiles':
            get_y_func = partial(_get_lbls, class_mapping=class_mapping)
        elif dataset_type == 'MultiLabeled_Tiles':
            get_y_func = _get_multi_lbls
        else:
            # Imagenet
            def get_y_func(x):
                return x.parent.stem
            if collate_fn is not _bb_pad_collate:
                databunch_kwargs['collate_fn'] = collate_fn
            else:
                databunch_kwargs['collate_fn'] = _ImagenetCollater(chip_size)
            _images_folder = os.path.join(os.path.abspath(path), 'images')
            if not os.path.exists(_images_folder):
                raise Exception(f"""Could not find a folder "images" in "{os.path.abspath(path)}",
                \na folder "images" should be present in the supplied path to work with "Imagenet" data_type. """
                )

        data = ArcGISImageList.from_folder(path / 'images') \
            .split_by_rand_pct(val_split_pct, seed=42) \
            .label_from_func(get_y_func)
        _show_batch_multispectral = show_batch_labeled_tiles

        if dataset_type == 'Imagenet':
            if class_mapping is None:
                class_mapping = {}
                index = 1
                for class_name in data.classes:
                    class_mapping[index] = class_name
                    index = index + 1

        if transforms is None:
            ranges = (0, 1)
            if _image_space_used == _map_space:
                train_tfms = [
                    rotate(degrees=30, p=0.5),
                    crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                    dihedral_affine(),
                    brightness(change=(0.4, 0.6)),
                    contrast(scale=(0.75, 1.5))
                ]
            else:
                train_tfms = [
                    rotate(degrees=30, p=0.5),
                    crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                    brightness(change=(0.4, 0.6)),
                    contrast(scale=(0.75, 1.5))
                ]
            val_tfms = [crop(size=chip_size, p=1.0, row_pct=0.5, col_pct=0.5)]
            transforms = (train_tfms, val_tfms)
    elif dataset_type == "superres" or dataset_type == "Export_Tiles":
        path_hr = path/'images'
        path_lr = path/'labels'
        il = ImageList.from_folder(path_hr)
        hr_suffix = il.items[0].suffix
        img_size = il[0].shape[1]
        downsample_factor = kwargs.get('downsample_factor', None)
        if downsample_factor is None:
            downsample_factor = 4
        path_lr_check = path/f'esri_superres_labels_downsample_factor.txt'
        prepare_label = False
        if path_lr_check.exists():
            with open(path_lr_check) as f:
                label_downsample_ratio = float(f.read())
            if label_downsample_ratio != downsample_factor:
                prepare_label = True
        else:
            prepare_label = True
        if prepare_label:
            parallel(partial(resize_one, path_lr=path_lr, size=img_size/downsample_factor, path_hr=path_hr, img_size=img_size), il.items, max_workers=databunch_kwargs.get('num_workers'))
            with open(path_lr_check, 'w') as f:
                f.write(str(downsample_factor))

        data = ImageImageListSR.from_folder(path_lr)\
            .split_by_rand_pct(val_split_pct, seed=seed)\
            .label_from_func(lambda x: path_hr/x.with_suffix(hr_suffix).name)
        if resize_to is None:
            kwargs_transforms['size'] = img_size
        kwargs_transforms['tfm_y'] = True
        
    elif dataset_type in ['ner_json','BIO','IOB','LBIOU','BILUO']:
        if batch_size == 64:
            batch_size = 8
        encoding = kwargs.get("encoding", "UTF-8")
        ner_architecture = kwargs.get("ner_architecture", "spacy")
        return _NERData(dataset_type=dataset_type, path=path, class_mapping=class_mapping, seed=seed,
                                val_split_pct=val_split_pct, batch_size=batch_size, encoding=encoding)
    elif dataset_type == "PointCloud":
        from ._utils.pointcloud_data import Transform3d
        if transforms is None:
            transform_fn = Transform3d()
        elif transforms is False:
            transform_fn = None
        else:
            transform_fn = transforms
        return pointcloud_prepare_data(path, class_mapping, batch_size, val_split_pct, dataset_type, transform_fn, **kwargs)
    elif dataset_type == "ImageCaptioning":
        from ._utils.image_captioning_data import prepare_captioning_dataset
        return prepare_captioning_dataset(path,
                                          chip_size,
                                          batch_size,
                                          val_split_pct,
                                          transforms,
                                          resize_to,
                                          **kwargs)
    elif dataset_type == "ChangeDetection":
        from ._utils.change_detection_data import prepare_change_detection_data
        kwargs.pop('rgb_bands', None)
        kwargs.pop('bands', None)
        kwargs.pop('norm_pct', None)
        return prepare_change_detection_data(path,
                                             chip_size,
                                             batch_size,
                                             val_split_pct,
                                             transforms,
                                             _is_multispectral=_is_multispectral,
                                             rgb_bands=rgb_bands,
                                             bands=bands,
                                             extract_bands=kwargs.pop('extract_bands', None),
                                             norm_pct=norm_pct,
                                             **kwargs)

    elif dataset_type == "CycleGAN":
        path = path/"Images"
        if _is_multispectral:
            data = prepare_data_ms_cyclegan(path, norm_pct, val_split_pct, seed, databunch_kwargs)
            data.n_channel = data.x[0].data[0].shape[0]
            data._is_multispectral = _is_multispectral
            data._imagery_type = _imagery_type
            data._bands = _bands
            data._norm_pct = norm_pct
            data._extract_bands = None
            data._do_normalize = False
            x_shape = data.train_ds[0][0].shape
            data.chip_size = x_shape[-1]
            return data
        data = ImageTupleList.from_folders(path, 'train_a', 'train_b')\
                .split_by_rand_pct(val_split_pct, seed=seed)\
                .label_empty()
        img_size = data.x[0].shape[-1]
        if resize_to is None:
            kwargs_transforms['size'] = img_size
    elif dataset_type == "Pix2Pix":
        path = path/"Images"
        if _is_multispectral:
            data = prepare_data_ms_pix2pix(path, norm_pct, val_split_pct, seed, databunch_kwargs)
            data.n_channel = data.x[0].data[0].shape[0]
            data._is_multispectral = _is_multispectral
            data._imagery_type = _imagery_type
            data._bands = _bands
            data._norm_pct = norm_pct
            data._extract_bands = None
            data._do_normalize = False
            x_shape = data.train_ds[0][0].shape
            data.chip_size = x_shape[-1]
            return data
        data = (ImageTupleList2.from_folders(path, 'train_a', 'train_b')
                      .split_by_rand_pct(val_split_pct, seed=seed)
                      .label_empty())
        img_size = data.x[0].shape[-1]
        if resize_to is None:
            kwargs_transforms['size'] = img_size
    else:
        raise NotImplementedError('Unknown dataset_type="{}".'.format(dataset_type))
    
    if _is_multispectral:
        # Normalize multispectral imagery by calculating stats
        if dataset_type == 'RCNN_Masks':
            kwargs['do_normalize'] = False

        data = (data.transform(transforms, **kwargs_transforms)
                    .databunch(**databunch_kwargs))

        if "InputRastersProps" in emd:
            # Starting with ArcGIS Pro 2.7 and Python API for ArcGIS 1.9, the following multispectral kwargs have been
            # deprecated. This is done in favour of the newly added support for Imagery statistics and metadata in the
            # IA > Export Training data for Deep Learining GP Tool.
            #
            #   bands, rgb_bands, norm_pct,
            #
            data._emd = emd
            data._sensor_name = emd['InputRastersProps']['SensorName']
            bands = data._band_names = emd['InputRastersProps']['BandNames']
            # data._band_mapping = {i: k for i, k in enumerate(bands)}
            # data._band_mapping_reverse = {k: i for i, k in data._band_mapping.items()}
            data._nbands = len(data._band_names)
            band_min_values = []
            band_max_values = []
            band_mean_values = []
            band_std_values = []
            for band_stats in  emd['AllTilesStats']:
                band_min_values.append(band_stats['Min'])
                band_max_values.append(band_stats['Max'])
                band_mean_values.append(band_stats['Mean'])
                band_std_values.append(band_stats['StdDev'])

            data._rgb_bands = rgb_bands
            data._symbology_rgb_bands = rgb_bands

            data._band_min_values = torch.tensor(band_min_values, dtype=torch.float32)
            data._band_max_values = torch.tensor(band_max_values, dtype=torch.float32)
            data._band_mean_values = torch.tensor(band_mean_values, dtype=torch.float32)
            data._band_std_values = torch.tensor(band_std_values, dtype=torch.float32)
            data._scaled_min_values = torch.zeros((data._nbands,), dtype=torch.float32)
            data._scaled_max_values = torch.ones((data._nbands,), dtype=torch.float32)
            data._scaled_mean_values = _tensor_scaler(data._band_mean_values, min_values=data._band_min_values, max_values=data._band_max_values, mode='minmax')
            data._scaled_std_values = data._band_std_values*(data._scaled_mean_values/data._band_mean_values)

            # Handover to next section
            norm_pct = 1
            bands = data._band_names
            rgb_bands = symbology_rgb_bands = sniff_rgb_bands(data._band_names)
            if rgb_bands is None:
                rgb_bands = []
                if len(data._band_names) < 3:
                    symbology_rgb_bands = [0] # Panchromatic
                else:
                    symbology_rgb_bands = [0, 1, 2] # Case where could not find RGB in multiband imagery
        else:
            symbology_rgb_bands = rgb_bands
            if len(data.x) < 300:
                norm_pct = 1

            # Statistics
            dummy_stats = {
                "batch_stats_for_norm_pct_0" : {
                    "band_min_values":None,
                    "band_max_values":None,
                    "band_mean_values":None,
                    "band_std_values":None,
                    "scaled_min_values":None,
                    "scaled_max_values":None,
                    "scaled_mean_values":None,
                    "scaled_std_values":None
                }
            }
            normstats_json_path = os.path.abspath(data.path / '..' / 'esri_normalization_stats.json')
            if not os.path.exists(normstats_json_path):
                normstats = dummy_stats
                with open(normstats_json_path, 'w', encoding='utf-8') as f:
                    json.dump(normstats, f, ensure_ascii=False, indent=4)
            else:
                with open(normstats_json_path) as f:
                    normstats = json.load(f)

            norm_pct_search = f"batch_stats_for_norm_pct_{round(norm_pct*100)}"
            if norm_pct_search in normstats:
                batch_stats = normstats[norm_pct_search]
                for s in batch_stats:
                    if batch_stats[s] is not None:
                        batch_stats[s] = torch.tensor(batch_stats[s])
            else:
                batch_stats = _get_batch_stats(data.x, norm_pct)
                normstats[norm_pct_search] = dict(batch_stats)
                for s in normstats[norm_pct_search]:
                    if normstats[norm_pct_search][s] is not None:
                        normstats[norm_pct_search][s] = normstats[norm_pct_search][s].tolist()
                with open(normstats_json_path, 'w', encoding='utf-8') as f:
                    json.dump(normstats, f, ensure_ascii=False, indent=4)

            # batch_stats -> [band_min_values, band_max_values, band_mean_values, band_std_values, scaled_min_values, scaled_max_values, scaled_mean_values, scaled_std_values]
            data._band_min_values = batch_stats['band_min_values']
            data._band_max_values = batch_stats['band_max_values']
            data._band_mean_values = batch_stats['band_mean_values']
            data._band_std_values = batch_stats['band_std_values']
            data._scaled_min_values = batch_stats['scaled_min_values']
            data._scaled_max_values = batch_stats['scaled_max_values']
            data._scaled_mean_values = batch_stats['scaled_mean_values']
            data._scaled_std_values = batch_stats['scaled_std_values']
        #

        # Prevent Divide by zeros
        data._band_max_values[data._band_min_values == data._band_max_values]+=1
        data._scaled_std_values[data._scaled_std_values == 0]+=1e-02
        
        # Scaling
        data._min_max_scaler = partial(_tensor_scaler, min_values=data._band_min_values, max_values=data._band_max_values, mode='minmax')
        data._min_max_scaler_tfm = partial(_tensor_scaler_tfm, min_values=data._band_min_values, max_values=data._band_max_values, mode='minmax')
        #data.add_tfm(data._min_max_scaler_tfm)
        
        # Transforms
        def _scaling_tfm(x): 
            ## Scales Fastai Image Scaling | MS Image Values -> 0 - 1 range
            return x.__class__(data._min_max_scaler_tfm((x.data,None))[0][0])
        
        ## Fastai need tfm, order and resolve.
        class dummy():
            pass
        _scaling_tfm.tfm = dummy()
        _scaling_tfm.tfm.order = 0
        _scaling_tfm.resolve = dummy

        ## Scaling the images before applying any  other transform
        if getattr(data.train_ds, 'tfms') is not None:
            data.train_ds.tfms = [_scaling_tfm] + data.train_ds.tfms
        else:
            data.train_ds.tfms = [_scaling_tfm]
        if getattr(data.valid_ds, 'tfms') is not None:
            data.valid_ds.tfms = [_scaling_tfm] + data.valid_ds.tfms
        else:
            data.valid_ds.tfms = [_scaling_tfm]

        # Normalize
        data._do_normalize = True
        if kwargs.get('do_normalize', None) is not None:
            data._do_normalize = kwargs.get('do_normalize', True)
        if data._do_normalize:
            data = data.normalize(stats=(data._scaled_mean_values, data._scaled_std_values), do_x=True, do_y=False)
        
    elif dataset_type == 'RCNN_Masks':
        data = (data.transform(transforms, **kwargs_transforms)
                .databunch(**databunch_kwargs))
        data.show_batch = types.MethodType( show_batch_rcnn_masks, data)
        # Exceptional case
        # We are dividing image pixel values by 255, at the time of opening it for rcnn masks
        # Not normalizing imagery here because model will normalize image internally
        data.train_ds.x._div = 255.
        data.valid_ds.x._div = 255.
    elif dataset_type == "Pix2Pix":
        data = (data.transform(get_transforms(), **kwargs_transforms)
            .databunch(**databunch_kwargs)) 
        data.n_channel = data.x[0].data[0].shape[0]
    
    elif dataset_type == "superres" or dataset_type == "Export_Tiles":
        data = (data.transform(get_transforms(), **kwargs_transforms)
            .databunch(**databunch_kwargs)
            .normalize(imagenet_stats, do_y=True))
    elif dataset_type == "CycleGAN":
        data = (data.transform(get_transforms(), **kwargs_transforms)
            .databunch(**databunch_kwargs))
        data.n_channel = data.x[0].data[0].shape[0]       
    else:
        data = (data.transform(transforms, **kwargs_transforms)
            .databunch(**databunch_kwargs)
            .normalize(imagenet_stats))
        # RGB Image
        # We need to divide image pixel values by 255. at the time of opening it
        # because same method is used to open multispectral imagery as well
        # and that workflow depends on the imagery specific stats
        # Inflating imagenet_stats by 255x should have also worked
        # But fastai transforms clip image value to 1 and
        # in fastai 1.0.60 transforms are applied before normalization
        data.train_ds.x._div = 255.
        data.valid_ds.x._div = 255.

    # Imagery type used while opening image chips
    data._imagery_type = imagery_type
    data.train_ds.x._imagery_type = data._imagery_type
    data.valid_ds.x._imagery_type = data._imagery_type

    # Assigning chip size from training dataset and not data.x 
    # to consider transforms and resizing
    x_shape = data.train_ds[0][0].shape
    data.chip_size = x_shape[-1]
    data._val_split_pct = val_split_pct

    # Alpha channel check with GDAL
    if HAS_GDAL and x_shape[0] == 4:
        if data._imagery_type == 'ASSUMED_RGB':
            message = f"""
            Could not infer Imagery Type, Found 4 Bands in input imagery. Please set the optional parameter 'imagery_type' to an appropriate value.
            \nIf the imagery used to export the training data is a RGB imagery, please continue training by specifying `imagery_type='RGB'`.
            \nIf the imagery used to export the training data is a multispectral imagery containing information in the 4th band, please check the documentation for parameter 'imagery_type' to find a suitable value. 
            """
            raise Exception(message)

    if has_esri_files:
        with open(stats_file) as f:
            stats = json.load(f)
            data._dataset_type = stats['MetaDataMode']
    else:
        data._dataset_type = dataset_type
    
    if dataset_type == "superres" or dataset_type == "Export_Tiles":
        data._dataset_type = "SuperResolution"

    if alter_class_mapping:
        new_mapping = {}
        for i, class_name in enumerate(class_mapping.keys()):
            new_mapping[i+1] = class_name
        class_mapping = new_mapping

    ## For calculating loss from inverse of frquency.
    if dataset_type == 'Classified_Tiles':
        with open(stats_file) as f:
            stats_json = json.load(f)
            pixel_stats = stats_json.get('ClassPixelStats', None)
            if pixel_stats is not None:
                data.num_pixels_per_class = pixel_stats.get('NumPixelsPerClass', None)
            else:
                data.num_pixels_per_class = None

           ## Might want to change the variable name
            if data.num_pixels_per_class is not None:
                num_pixels_per_class = np.array(data.num_pixels_per_class, dtype=np.int64)
                if num_pixels_per_class.sum() < 0:
                    data.overflow_encountered = True
                    data.class_weight = None
                else:
                    _num_pixels_per_class = np.copy(num_pixels_per_class)
                    _num_pixels_per_class[_num_pixels_per_class==0]=1
                    data.class_weight = num_pixels_per_class.sum() / _num_pixels_per_class
                    data.class_weight[num_pixels_per_class==0]=0
            else:
                data.class_weight = None

    data.class_mapping = class_mapping
    data.color_mapping = color_mapping
    data.show_batch = types.MethodType(
        types.FunctionType(
            data.show_batch.__code__, data.show_batch.__globals__, data.show_batch.__name__,
            (min(int(math.sqrt(data.batch_size)), 5), *data.show_batch.__defaults__[1:]), data.show_batch.__closure__
        ),
        data
    )
    data.orig_path = path
    data.resize_to = kwargs_transforms.get('size', None)
    data.height_width = height_width
    data.downsample_factor = kwargs.get("downsample_factor")
    data.dataset_type = dataset_type
    
    data._is_multispectral = _is_multispectral
    if data._is_multispectral:
        data._bands = bands
        data._norm_pct = norm_pct
        data._rgb_bands = rgb_bands
        data._symbology_rgb_bands = symbology_rgb_bands

        # Handle invalid color mapping 
        data._multispectral_color_mapping = color_mapping
        if any( -1 in x for x in data._multispectral_color_mapping.values() ):
            random_color_list = np.random.randint(low=0, high=255, size=(len(data._multispectral_color_mapping), 3)).tolist()
            for i, c in enumerate(data._multispectral_color_mapping):
                if -1 in data._multispectral_color_mapping[c]:
                    data._multispectral_color_mapping[c] = random_color_list[i]

        # prepare color array
        alpha = kwargs.get('alpha', 0.7)
        color_array = torch.tensor( list(data.color_mapping.values()) ).float() / 255
        alpha_tensor = torch.tensor( [alpha]*len(color_array) ).view(-1, 1).float()
        color_array = torch.cat( [ color_array, alpha_tensor ], dim=-1)
        background_color = torch.tensor( [[0, 0, 0, 0]] ).float()
        data._multispectral_color_array = torch.cat( [background_color, color_array] )

        # Prepare unknown bands list if bands data is missing
        if data._bands is None:
            n_bands = data.x[0].data.shape[0]
            if n_bands == 1:# Handle Pancromatic case
                data._bands = ['p']
                data._symbology_rgb_bands = [0]
            else:
                data._bands = ['u' for i in range(n_bands)]
                if n_bands == 2:# Handle Data with two channels
                    data._symbology_rgb_bands = [0]
        
        # 
        if data._rgb_bands is None:
            data._rgb_bands = []

        # 
        if data._symbology_rgb_bands is None:
            data._symbology_rgb_bands = [0, 1, 2][:min(n_bands, 3)]

        # Complete symbology rgb bands
        if len(data._bands) > 2 and len(data._symbology_rgb_bands) < 3:
            data._symbology_rgb_bands += [ min(max(data._symbology_rgb_bands)+1, len(data._bands)-1)  for i in range(3 - len(data._symbology_rgb_bands)) ]
        
        # Overwrite band values at r g b indexes with 'r' 'g' 'b'
        for i, band_idx in enumerate(data._rgb_bands):
            if band_idx is not None:
                if data._bands[band_idx] == 'u':
                    data._bands[band_idx] = ['r', 'g', 'b'][i]

        # Attach custom show batch
        if _show_batch_multispectral is not None:
            data.show_batch = types.MethodType( _show_batch_multispectral, data )

        # Apply filter band transformation if user has specified extract_bands otherwise add a generic extract_bands
        """
        extract_bands : List containing band indices of the bands from imagery on which the model would be trained. 
                        Useful for benchmarking and applied training, for reference see examples below.
                        
                        4 band naip ['r, 'g', 'b', 'nir'] + extract_bands=[0, 1, 2] -> 3 band naip with bands ['r', 'g', 'b'] 

        """
        data._extract_bands = kwargs.get('extract_bands', None) 
        if data._extract_bands is None:
            data._extract_bands = list(range(len(data._bands)))
        else:
            data._extract_bands_tfm = partial(_extract_bands_tfm, band_indices=data._extract_bands)
            data.add_tfm(data._extract_bands_tfm)   

        # Tail Training Override
        _train_tail = True
        if [data._bands[i] for i in data._extract_bands] == ['r', 'g', 'b']:
            _train_tail = False
        data._train_tail = kwargs.get('train_tail', _train_tail)

    if not_label_count[0]:
            logger = logging.getLogger()
            logger.warning("Please check your dataset. " + str(not_label_count[0]) + " images dont have the corresponding label files.")

    data._image_space_used = _image_space_used

    return data


"""
Connection Object that uses Python Requests

Requires: requests, requests_toolbelt,
Possible optional might be required: requests_ntlm, requests_kerberos, requests-oauthlib

"""
try:
    import arcpy
    HASARCPY = True
except ImportError:
    HASARCPY = False
except:
    HASARCPY = False

import sys
if sys.platform == 'win32':
    try:
        import certifi_win32
    except ImportError:
        pass

import os
import copy
import json
import uuid
import datetime
import mimetypes
import tempfile
from urllib.request import urlparse, unquote, urljoin
import requests
from requests import Session
from requests_toolbelt.downloadutils import stream
from requests_toolbelt.multipart.encoder import MultipartEncoder
from ._helpers import _filename_from_headers, _filename_from_url
from ._authguess import GuessAuth
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._isd import InsensitiveDict

__version__ = "1.8.4"

_DEFAULT_TOKEN = uuid.uuid4()

class Connection(object):
    """
    Universal Connection Object
    """
    _refresh_token = None
    _token = None
    _token_url = None
    _create_time = None
    _session = None
    _header = None
    _username = None
    _password = None
    _proxy_url = None
    _proxy_port = None
    _proxy_password = None
    _proxy_username = None
    _verify_cert = None
    _baseurl = None
    baseurl = None
    _referer = None
    _expiration = None
    _auth = None
    _product = None
    _custom_auth = None
    #----------------------------------------------------------------------
    def __init__(self,
                 baseurl=None,
                 username=None,
                 password=None,
                 **kwargs):
        """initializer

        Optional Kwargs

        client_id=None
        client_secret=None
        expiration=60
        cert_file
        key_file
        proxy_url
        proxy_port
        verify_cert
        referer
        all_ssl = True
        portal_connection = None
        product = "UNKNOWN", "PORTAL", "SERVER", "AGOL", "GEOEVENT"  # SHOULD BE SET BY GIS object, or Server Object, etc...
        token= None If token, _AUTH = BUILTIN
        token_url
        AUTH keys = HOME, BUILTIN, PRO, ANON, PKI, HANDLER, UNKNOWN (Internal)
        custom_auth = Requests authencation handler
        trust_env = T/F if to ignore netrc files
        """
        from arcgis.gis import GIS
        self._all_ssl = kwargs.pop("all_ssl", True)
        self.trust_env = kwargs.pop("trust_env", None)
        if baseurl:
            while baseurl.endswith("/"):
                baseurl = baseurl[:-1]
        self._custom_auth = kwargs.pop('custom_auth', None)
        self._product = kwargs.pop('product', "UNKNOWN")
        if baseurl is None:
            self._product = "AGOL"
            baseurl = "https://www.arcgis.com"


        self._baseurl = baseurl
        self._username = username
        self._password = password

        self._expiration = kwargs.pop('expiration', 60) or 60
        self._portal_connection = kwargs.pop('portal_connection', None) # For Federated Objects (Portal Connection)
        if isinstance(self._portal_connection, GIS):
            self._portal_connection = self._portal_connection._con

        if (self._referer or self._referer is None) and \
           self._portal_connection and \
           str(self._portal_connection._auth).lower() == "home":
            self._referer = None
        elif baseurl.lower() == 'pro':
            try:
                self._referer = arcpy.GetSigninToken().pop('referer', 'http')
            except:
                self._referer = kwargs.pop('referer', 'http')
        else:
            self._referer = kwargs.pop('referer', 'http')

        self._verify_cert = kwargs.pop("verify_cert", False)#True)
        if self._verify_cert == False:
            import warnings
            from requests.packages.urllib3 import exceptions
            warnings.simplefilter("ignore", exceptions.InsecureRequestWarning)

        self._cert_file = kwargs.pop('cert_file', None)
        self._key_file = kwargs.pop('key_file', None)
        if 'proxy_host' in kwargs:
            self._proxy_url = kwargs.pop('proxy_host', None)
        else:
            self._proxy_url = kwargs.pop('proxy_url', None)
        self._proxy_port = kwargs.pop('proxy_port', None)
        self._proxy_username = kwargs.pop('proxy_username', None)
        self._proxy_password = kwargs.pop('proxy_password', None)
        self._header = {'User-Agent' : 'Geosaurus/%s' % __version__}
        self._token = kwargs.pop('token', None)
        self._client_id = kwargs.pop('client_id', None)
        self._client_secret = kwargs.pop('client_secret', None)
        self._token_url = kwargs.pop('token_url', None)

        if username is None and password is None and self._portal_connection is None:
            self._auth = "ANON"
        elif ((not username is None and not password is None) or self._portal_connection):
            self._auth = "BUILTIN"
        if (username and password) and \
           self._client_id is None and \
           str(baseurl).lower() != "pro":
            self._auth = "BUILTIN"
        elif self._portal_connection:
            self._auth = "BUILTIN"
        elif baseurl.lower() == 'pro':
            self._auth = "PRO"
            self._baseurl = arcpy.GetActivePortalURL()
        elif self._cert_file or\
             (self._cert_file and self._key_file):
            self._auth = "PKI"

        self._create_session()

        #  Product Info
        if self._token:
            if self._expiration is None:
                self._expiration = 60
            self._create_time = datetime.datetime.now() + datetime.timedelta(minutes=self._expiration)
            self._auth = "BUILTIN"
        elif self._client_id:
            self._product = self._check_product()
            if self._product in ['PORTAL', "AGOL"]:
                resp = self.post("/portals/self", {'f' : 'json'}, add_token=False)
                issaml = resp.get("samlEnabled", False)
                isoauth = resp.get("supportsOAuth", False)
            else:
                resp = None
                issaml = False
                isoauth = False
            self._auth = 'OAUTH'
            parsed = urlparse(self._baseurl)
            wa = parsed.path
            if wa.startswith('/'):
                wa = wa[1:].split('/')[0]
            else:
                wa = wa.split('/')[0]
            if len(wa) > 0:
                self._token_url = "https://%s/%s/sharing/rest/oauth2/token" % (parsed.netloc, wa)
            else:
                self._token_url = "https://%s/sharing/rest/oauth2/token" % (parsed.netloc)

        else:
            self._product = self._check_product()
        self._baseurl = self._validate_url(self._baseurl)
        self.baseurl = self._baseurl
    #----------------------------------------------------------------------
    def _validate_url(self, url):
        """ensures the base url has the /sharing/rest"""
        if url.lower().find('arcgis.com') > -1:
            self._product = "AGOL"
        if self._product in ['AGO', 'AGOL', 'PORTAL']:
            if not url[-1] == '/':
                url += '/'
            if url.lower().find("www.arcgis.com") > -1:
                urlscheme = urlparse(url).scheme
                return "{scheme}://www.arcgis.com/sharing/rest/".format(scheme=urlscheme)
            elif url.lower().endswith("sharing/"):
                return url + 'rest/'
            elif url.lower().endswith("sharing/rest/"):
                return url
            else:
                return url + 'sharing/rest/'
        elif self._product in ['SERVER', 'FEDERATED_SERVER', "FEDERATED"]:
            parsed = urlparse(url)
            path = parsed.path
            if str(path).startswith("/"):
                path = path[1:]
            url = "%s://%s/%s/rest/services/" % (parsed.scheme,
                                                 parsed.netloc,
                                                 path.split('/')[0])
        return url
    #----------------------------------------------------------------------
    def _create_session(self):
        if self._proxy_port and self._proxy_url:
            url = "%s:%s" % (self._proxy_url, self._proxy_port)
            if self._proxy_password and self._proxy_username:
                proxies = {'http': 'http://%s:%s@%s' % (self._proxy_username,
                                                        self._proxy_password,
                                                        url),
                           'https': 'https://%s:%s@%s' % (self._proxy_username,
                                                          self._proxy_password,
                                                          url)}
            else:
                proxies = {'http': 'http://%s' % url,
                           'https': 'https://%s' % url}
        else:
            proxies = None

        self._session = Session()
        self._session.verify = self._verify_cert
        self._session.stream = True
        self._session.trust_env = self.trust_env
        self._session.headers.update(self._header)
        self._session.proxies = proxies
        if self._referer is None and\
           (self._portal_connection and \
           str(self._portal_connection._auth).lower() == "home"):
            self._referer = "http"
            self._session.headers.pop("Referer", None)
            self._session.headers['Referer'] = json.dumps("")
        elif (self._portal_connection and str(self._portal_connection._auth).lower() == "home"):
            self._referer = "http"
            self._session.headers.pop("Referer", None)
            self._session.headers['Referer'] = json.dumps("")
        elif self._referer is None:
            self._referer = 'http'
            self._session.headers.update({'Referer': self._referer})
        else:
            self._session.headers.update({'Referer': self._referer})
        if self._custom_auth:
            self._session.auth = self._custom_auth
            self._auth = "CUSTOM"
        elif self._username and self._password:
            self._session.auth = GuessAuth(username=self._username,
                                           password=self._password)
        elif self._auth.lower() == "pro":
            self._session.auth = GuessAuth(None, None)
        else:
            try:
                from requests_negotiate_sspi import HttpNegotiateAuth
                HAS_KERBEROS = True
            except:
                HAS_KERBEROS = False
            if HAS_KERBEROS:
                self._session.auth = HttpNegotiateAuth()
            else:
                try:
                    from requests_kerberos import HTTPKerberosAuth, DISABLED
                    self._session.auth = HTTPKerberosAuth(DISABLED)
                except ImportError:
                    pass
                except Exception as e:
                    raise e
        if self._cert_file and self._key_file:
            self._session.cert = (self._cert_file,
                                  self._key_file)
        elif self._cert_file and self._password:
            from arcgis.gis._impl._con._cert import pfx_to_pem
            self._key_file, self._cert_file = pfx_to_pem(pfx_path=self._cert_file, pfx_password=self._password)
            self._session.cert = (self._cert_file, self._key_file)
        elif self._cert_file:
            self._session.cert = (self._cert_file)
        else:
            self._session.cert = None
    #----------------------------------------------------------------------
    def get(self,
            path,
            params=None,
            **kwargs):
        """

        sends a GET request.

        ===========================   =====================================================
        **optional keywords**         **description**
        ---------------------------   -----------------------------------------------------
        path                          required string. URI resource to access.
        ===========================   =====================================================

        ===========================   =====================================================
        **optional keywords**         **description**
        ---------------------------   -----------------------------------------------------
        params                        optional dictionary.  This is the data payload to the
                                      web resource
        ---------------------------   -----------------------------------------------------
        out_folder                    optional string. Save path on disk.  The default is
                                      the user's temp directory.
        ---------------------------   -----------------------------------------------------
        file_name                     optional string. Name of the file to save. If not
                                      provided, name will try to be parsed from the response.
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If True, the marker f='json' will
                                      be appended to the parameters.
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean. Default True, if true, the call will
                                      add a token to any token based security.
        ---------------------------   -----------------------------------------------------
        json_encode                   optional Boolean.  When False, the JSON values will not be encoded.
        ===========================   =====================================================
        """
        json_encode = kwargs.pop('json_encode', True)
        if self._baseurl.endswith('/') == False:
            self._baseurl += "/"
        url = path
        token = kwargs.pop('token', _DEFAULT_TOKEN)
        token_as_header = kwargs.pop('token_as_header', False)
        token_header = kwargs.pop('token_header', "X-Esri-Authorization")
        if url.find('://') == -1:
            if url.startswith('/') == False and \
               self._baseurl.endswith('/') == False:
                url = "/" + url
            url = self._baseurl + url
        if kwargs.pop("ssl", False) or self._all_ssl:
            url = url.replace("http://", "https://")
        if params is None:
            params = {}
        #if self._auth == "IWA":
        #    self._session = None
        if self._session is None:
            self._create_session()

        try_json = kwargs.pop("try_json", True)
        add_token = kwargs.pop('add_token', True)

        if add_token:
            if token != _DEFAULT_TOKEN:
                if token is not None:
                    params['token'] = token
                else:
                    params.pop('token', None)
                    #pass
            elif token_as_header == False and self.token is not None: #as ?token=
                params['token'] = self.token
            elif token_as_header and self.token is not None:#(token and token != _DEFAULT_TOKEN): # as X-Esri-Auth header with given token
                self._session.headers.update({token_header: "Bearer %s" % token})
            elif token_as_header and token_header and self.token: # as X-Esri-Auth header with generated token
                self._session.headers.update({token_header: "Bearer %s" % self.token})
        if try_json:
            params['f'] = 'json'
        if params == {}:
            params = None

        if 'out_folder' in kwargs:
            out_path = kwargs.pop('out_folder',
                                  tempfile.gettempdir())
        else:
            out_path = kwargs.pop('out_path',
                                  tempfile.gettempdir())
        file_name = kwargs.pop('file_name', None)
        if params and json_encode:
            for k, v in copy.copy(params).items():
                if isinstance(v, (tuple, dict, list, bool)):
                    params[k] = json.dumps(v)
                elif isinstance(v, PropertyMap):
                    params[k] = json.dumps(dict(v))
                elif isinstance(v, InsensitiveDict):
                    params[k] = v.json
        try:
            if self._cert_file:
                cert = (self._cert_file, self._key_file)
            else:
                cert = None
            resp = self._session.get(url=url,
                                     params=params,
                                     cert=cert,
                                     verify=self._verify_cert)

        except requests.exceptions.SSLError as err:
            raise requests.exceptions.SSLError(
                "Please set verify_cert=False due to encountered SSL error: %s" % err)
        except requests.exceptions.InvalidURL as errIU:
            raise requests.exceptions.SSLError(
                "Invalid URL provided: %s" % errIU)
        except requests.exceptions.ConnectionError as errCE:
            raise requests.exceptions.ConnectionError(
                "A connection error has occurred: %s" % errCE)
        except requests.exceptions.InvalidHeader as errIH:
            raise requests.exceptions.InvalidHeader(
                "A invalid header was provided: %s" % errIH)
        except requests.exceptions.HTTPError as errh:
            raise requests.exceptions.HTTPError(
                "Http Error: %s" % errh)
        except requests.exceptions.MissingSchema as errMS:
            raise requests.exceptions.MissingSchema(
                "URL scheme must be provided: %s" % errMS)
        except requests.exceptions.RequestException as errRE:
            raise requests.exceptions.RequestException(
                "A general expection was raised: %s"  % errRE)
        except Exception as e:
            raise Exception('A general error occurred: %s' % e)
        except:
            import traceback
            raise Exception(
                'An unknown error occurred: %s' % traceback.format_exc())

        return self._handle_response(resp, file_name, out_path, try_json, force_bytes=kwargs.pop('force_bytes', False))
    #----------------------------------------------------------------------
    def _handle_response(self, resp, file_name, out_path, try_json, force_bytes=False):
        """
        handles the request responses

        ===========================   =====================================================
        **optional keywords**         **description**
        ---------------------------   -----------------------------------------------------
        resp                          required resquests.Rsponse object.
        ---------------------------   -----------------------------------------------------
        file_name                     required string.  Name of the output file if needed.
        ---------------------------   -----------------------------------------------------
        out_path                      required string. Name of the save folder.
        ---------------------------   -----------------------------------------------------
        try_json                      required boolean. Determines if the response should
                                      be returned as a dictionary or the native format
                                      send back by a server
        ---------------------------   -----------------------------------------------------
        force_bytes                   Optional Boolean. If True, the results returns as bytes
                                      instead of a file path.
        ===========================   =====================================================

        returns: string, dictionary, or bytes depending on the response.

        """
        if 'Set-Cookie' in resp.headers and \
           (str(self._auth).lower() in ['anon'] or \
            (str(self._username).find("\\") > -1 and str(self._auth).lower() in ['builtin'])):
            self._auth = 'IWA'

        data = None
        url = resp.url
        if out_path and \
           os.path.isdir(out_path) == False:
            os.makedirs(out_path)
        if out_path is None:
            out_path = tempfile.gettempdir()
        if file_name is None and \
           (resp.headers['Content-Type'].lower().find('json') == -1 and \
           resp.headers['Content-Type'].lower().find('text') == -1):
            file_name = _filename_from_url(url) or _filename_from_headers(
                resp.headers) or None
        if force_bytes:
            try:
                return bytes(resp.content)
            except:
                return resp.content
        if file_name is not None:
            file_name = os.path.join(out_path, file_name)
            if os.path.isfile(file_name):
                os.remove(file_name)
            fp = stream.stream_response_to_file(response=resp,
                                                path=file_name,
                                                chunksize=512*2)
            return fp

        if try_json:
            if 'Content-Length' in resp.headers and int(resp.headers.get('Content-Length')) == 0:
                data = {}
            elif 'Transfer-Encoding' in resp.headers and \
               resp.headers['Transfer-Encoding'].lower() == 'chunked':
                data = None
                for it in resp.iter_lines(chunk_size=None, decode_unicode=True, delimiter=None):
                    if data is None:
                        data = it
                    else:
                        data += it
                data = json.loads(data)
                if 'error' in data:
                    raise Exception(data['error'])
            else:
                data = resp.json()
            #if 'error' in data:
                #raise Exception(data['error'])
            #return data
        #else:
            #return resp.text
            if 'error' in data:
                if 'messages' in data:
                    return data
                errorcode = data['error']['code'] if 'code' in data['error'] else 0
                self._handle_json_error(data['error'], errorcode)
            return data
        else:
            return resp.text
    #----------------------------------------------------------------------
    def _handle_json_error(self, error, errorcode):
        errormessage = error.get('message')
        # handles case where message exists in the dictionary but is None
        if errormessage is None:
            errormessage = "Unknown Error"
        #_log.error(errormessage)
        if 'details' in error and error['details'] is not None:
            if isinstance(error['details'], str):
                errormessage = f"{errormessage} \n {error['details']}"
                #_log.error(error['details'])
            else:
                for errordetail in error['details']:
                    if isinstance(errordetail, str):
                        errormessage = errormessage + "\n" + errordetail
                        #_log.error(errordetail)

        errormessage = errormessage + "\n(Error Code: " + str(errorcode) +")"
        raise Exception(errormessage)
    #----------------------------------------------------------------------
    def post(self,
             path,
             params=None,
             files=None,
             **kwargs):
        """
        sends a POST request.

        ===========================   =====================================================
        **Parameters**                **Description**
        ---------------------------   -----------------------------------------------------
        path                          optional string.  URL or part of the url resource
                                      to call.
        ---------------------------   -----------------------------------------------------
        params                        optional dict.  Contains data to pass to the web
                                      resource.
        ---------------------------   -----------------------------------------------------
        files                         optional list of files

                                      Files can be provided two ways:

                                      The most basic way is:
                                      Way1: {key : r"c:\temp\myfile.foo}
                                      This is just the file path and the key.


                                      The preferred way:

                                      Way 2: {key : (file_name, open(c:\temp\myfile.foo, 'rb'), image\jpeg)}

                                      Way 2 requires providing the filename, IO object as 'rb', and the mimetype.
        ===========================   =====================================================


        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean.  True means try to add the boolean,
                                      else do not add a ?token=<foo> to the call. If
                                      is_geoevent is True, then the token will be appended
                                      to the header.
        ---------------------------   -----------------------------------------------------
        is_geoevent                   optional boolean. True means the auth token will be placed in the header
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If true, the call adds the ?f=json.
        ---------------------------   -----------------------------------------------------
        ssl                           optional boolean. If true all calls are forced to be
                                      https.
        ---------------------------   -----------------------------------------------------
        out_folder                    optional string.  This is the save folder for the data.
        ---------------------------   -----------------------------------------------------
        file_name                     optional string. The save name of the file. This will override the file name if provided in the response.
        ---------------------------   -----------------------------------------------------
        force_bytes                   optional boolean.  Deprecated.
        ---------------------------   -----------------------------------------------------
        add_headers                   optional dict.  If provided, additional headers will be given for a single call.
        ---------------------------   -----------------------------------------------------
        post_json                     optional bool. If True, the data is pushed in the request's json parameter.  This is an edge case for Workflow Manager. The default is `False`.
        ---------------------------   -----------------------------------------------------
        json_encode                   optional Bool. If False, the key/value parameters will not be JSON encoded.
        ===========================   =====================================================

        :returns: data returned from the URL call.

        """
        json_encode = kwargs.pop("json_encode", True)
        if self._baseurl.endswith("/") == False:
            self._baseurl += "/"
        url = path
        token = kwargs.pop('token', _DEFAULT_TOKEN)
        post_json = kwargs.pop("post_json", False)
        token_as_header = kwargs.pop('token_as_header', False)
        token_header = kwargs.pop('token_header', "X-Esri-Authorization")
        #if self._auth == "IWA":
        #    self._session = None
        if 'postdata' in kwargs: # handles legacy issues
            params = kwargs.pop('postdata')
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()
        try_json = kwargs.pop("try_json", True)
        add_token = kwargs.pop('add_token', True)
        if url.find('://') == -1:
            url = self._baseurl + url
        if kwargs.pop("ssl", False) or self._all_ssl:
            url = url.replace("http://", "https://")
        if add_token:
            if token != _DEFAULT_TOKEN:
                if token is not None:
                    params['token'] = token
                else:
                    params.pop('token', None)
                    #pass
            elif token_as_header == False and self.token is not None: #as ?token=
                params['token'] = self.token
            elif token_as_header and self.token is not None:#(token and token != _DEFAULT_TOKEN): # as X-Esri-Auth header with given token
                self._session.headers.update({token_header: "Bearer %s" % token})
            elif token_as_header and token_header and self.token: # as X-Esri-Auth header with generated token
                self._session.headers.update({token_header: "Bearer %s" % self.token})

        if try_json:
            params['f'] = 'json'
        if files:
            fields = { }
            if isinstance(files, dict):
                for k, v in files.items():
                    if isinstance(v, (list, tuple)):
                        fields[k] = v
                    else:
                        fields[k] = (os.path.basename(v),
                                     open(v, 'rb'),
                                     mimetypes.guess_type(v)[0])
            elif isinstance(files, (list, tuple)):
                for key, filePath, fileName in files:
                    if isinstance(fileName, str):
                        fields[key] = (fileName,
                                       open(filePath, 'rb'),
                                       mimetypes.guess_type(filePath)[0])
                    else:
                        fields[key] = v
            files = fields

        out_path = kwargs.pop('out_path', None) or \
            kwargs.pop('out_folder', None) or \
            tempfile.gettempdir()
        file_name = kwargs.pop('file_name', None)
        try:
            if self._cert_file:
                cert = (self._cert_file, self._key_file)
            else:
                cert = None
            if json_encode:
                for k,v in params.items():
                    if isinstance(v, (dict, list, tuple, bool)):
                        params[k] = json.dumps(v)
                    elif isinstance(v, PropertyMap):
                        params[k] = json.dumps(dict(v))
                    elif isinstance(v, InsensitiveDict):
                        params[k] = v.json
            if post_json:  # edge case workflow
                resp = self._session.post(url=url,
                                          json=params,
                                          cert=cert,
                                          files=files)
            else:
                resp = self._session.post(url=url,
                                          data=params,
                                          cert=cert,
                                          files=files)
        except requests.exceptions.SSLError as err:
            raise requests.exceptions.SSLError(
                "Please set verify_cert=False due to encountered SSL error: %s" % err)
        except requests.exceptions.InvalidURL as errIU:
            raise requests.exceptions.SSLError(
                "Invalid URL provided: %s" % errIU)
        except requests.exceptions.ConnectionError as errCE:
            raise requests.exceptions.ConnectionError(
                "A connection error has occurred: %s" % errCE)
        except requests.exceptions.InvalidHeader as errIH:
            raise requests.exceptions.InvalidHeader(
                "A invalid header was provided: %s" % errIH)
        except requests.exceptions.HTTPError as errh:
            raise requests.exceptions.HTTPError(
                "Http Error: %s" % errh)
        except requests.exceptions.RequestException as errRE:
            raise requests.exceptions.RequestException(
                "A general expection was raised: %s"  % errRE)
        except requests.exceptions.MissingSchema as errMS:
            raise requests.exceptions.MissingSchema(
                "URL scheme must be provided: %s" % errMS)
        except Exception as e:
            raise Exception('A general error occurred: %s' % e)
        except:
            import traceback
            raise Exception(
                'An unknown error occurred: %s' % traceback.format_exc())

        return self._handle_response(resp=resp,
                                     out_path=out_path,
                                     file_name=file_name,
                                     try_json=try_json,
                                     force_bytes=kwargs.pop('force_bytes', False))
    #----------------------------------------------------------------------
    def put(self, url, params=None, files=None, **kwargs):
        """
        sends a PUT request

        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        url                           Optional String. The web endpoint.
        ---------------------------   -----------------------------------------------------
        params                        Optional dict. The Key/value pairs to send along with the request.
        ---------------------------   -----------------------------------------------------
        files                         Optional list. Allows users to provide a file or files
                                      to the operation.  The format should be:
                                      files = [[key, filePath, fileName], ...,[keyN, filePathN, fileNameN]]
        ===========================   =====================================================

        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean.  True means try to add the boolean,
                                      else do not add a ?token=<foo> to the call.
        ---------------------------   -----------------------------------------------------
        token_as_header               Optional boolean.  If True, the token will go into
                                      the header as `Authorization` header.  This can be
                                      overwritten using the `token_header` parameter
        ---------------------------   -----------------------------------------------------
        token_header                  Optional String. If provided and token_as_header is
                                      True, authentication token will be placed in this
                                      instead on URL string.
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If true, the call adds the ?f=json.
        ---------------------------   -----------------------------------------------------
        ssl                           optional boolean. If true all calls are forced to be
                                      https.
        ---------------------------   -----------------------------------------------------
        post_json                     optional bool. If True, the data is pushed in the request's json parameter.  This is an edge case for Workflow Manager. The default is `False`.
        ---------------------------   -----------------------------------------------------
        json_encode                   optional Bool. If False, the key/value parameters will not be JSON encoded.
        ===========================   =====================================================

        :returns: dict or string depending on the response

        """
        token = kwargs.pop('token', _DEFAULT_TOKEN)
        post_json = kwargs.pop("post_json", False)
        json_encode = kwargs.pop("json_encode", True)
        out_path = kwargs.pop('out_path', None)
        file_name = kwargs.pop('file_name', None)
        token_as_header = kwargs.pop('token_as_header', True)
        token_header = kwargs.pop('token_header', "X-Esri-Authorization")
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()
        try_json = kwargs.pop("try_json", True)
        add_token = kwargs.pop('add_token', True)
        if url.find('://') == -1:
            url = self._baseurl + url
        if kwargs.pop("ssl", False):
            url = url.replace("http://", "https://")
        if add_token:
            if token != _DEFAULT_TOKEN:
                if token is not None:
                    params['token'] = token
                else:
                    params.pop('token', None)
                    #pass
            elif token_as_header == False and self.token is not None: #as ?token=
                params['token'] = self.token
            elif token_as_header and self.token is not None:#(token and token != _DEFAULT_TOKEN): # as X-Esri-Auth header with given token
                self._session.headers.update({token_header: "Bearer %s" % token})
            elif token_as_header and token_header and self.token: # as X-Esri-Auth header with generated token
                self._session.headers.update({token_header: "Bearer %s" % self.token})

        if try_json:
            params['f'] = 'json'

        if files:
            fields = { }
            if isinstance(files, dict):
                for k, v in files.items():
                    if isinstance(v, (list, tuple)):
                        fields[k] = v
                    else:
                        fields[k] = (os.path.basename(v),
                                     open(v, 'rb'),
                                     mimetypes.guess_type(v)[0])
            elif isinstance(files, (list, tuple)):
                for key, filePath, fileName in files:
                    if isinstance(fileName, str):
                        fields[key] = (fileName,
                                       open(filePath, 'rb'),
                                       mimetypes.guess_type(filePath)[0])
                    else:
                        fields[key] = v

            params.update(fields)
            del files, fields
        if self._cert_file:
            cert = (self._cert_file, self._key_file)
        else:
            cert = None
        if json_encode:
            for k,v in params.items():
                if isinstance(v, (dict, list, tuple, bool)):
                    params[k] = json.dumps(v)
                elif isinstance(v, PropertyMap):
                    params[k] = json.dumps(dict(v))
                elif isinstance(v, InsensitiveDict):
                    params[k] = v.json
        if post_json:  # edge case workflow
            resp = self._session.put(url=url,
                                      json=params,
                                      cert=cert,
                                      files=files)
        else:
            resp = self._session.put(url=url,
                                     data=params,
                                     cert=cert,
                                     files=files)
        #
        return self._handle_response(resp=resp,
                                         out_path=out_path,
                                         file_name=file_name,
                                         try_json=try_json,
                                         force_bytes=kwargs.pop('force_bytes', False))
    #----------------------------------------------------------------------
    def delete(self, url, params=None, **kwargs):
        """
        sends a PUT request

        ===========================   =====================================================
        **Parameters**                **Description**
        ---------------------------   -----------------------------------------------------
        url                           Optional String. The web endpoint.
        ---------------------------   -----------------------------------------------------
        params                        Optional dict. The Key/value pairs to send along with the request.
        ===========================   =====================================================

        ===========================   =====================================================
        **Optional Parameters**       **Description**
        ---------------------------   -----------------------------------------------------
        add_token                     optional boolean.  True means try to add the boolean,
                                      else do not add a ?token=<foo> to the call.
        ---------------------------   -----------------------------------------------------
        token_as_header               Optional boolean.  If True, the token will go into
                                      the header as `Authorization` header.  This can be
                                      overwritten using the `token_header` parameter
        ---------------------------   -----------------------------------------------------
        token_header                  Optional String. If provided and token_as_header is
                                      True, authentication token will be placed in this
                                      instead on URL string.
        ---------------------------   -----------------------------------------------------
        try_json                      optional boolean.  If true, the call adds the ?f=json.
        ---------------------------   -----------------------------------------------------
        ssl                           optional boolean. If true all calls are forced to be
                                      https.
        ===========================   =====================================================

        :returns: dict or string depending on the response
        """
        out_path = kwargs.pop('out_path', None)
        file_name = kwargs.pop('file_name', None)
        token_as_header = kwargs.pop('token_as_header', True)
        token_header = kwargs.pop('token_header', "X-Esri-Authorization")
        if params is None:
            params = {}
        if self._session is None:
            self._create_session()
        try_json = kwargs.pop("try_json", True)
        add_token = kwargs.pop('add_token', True)
        if url.find('://') == -1:
            url = self._baseurl + url
        if kwargs.pop("ssl", False):
            url = url.replace("http://", "https://")
        if add_token:
            if token_as_header == False and not 'token' in kwargs: #as ?token=
                params['token'] = self.token
            elif token_as_header == False and 'token' in kwargs: #as ?token= and user provides the token
                params['token'] = kwargs['token']
            elif token_as_header and 'token' in kwargs: # as X-Esri-Auth header with given token
                self._session.headers.update({token_header: "Bearer %s" % kwargs['token']})
            elif token_as_header and token_header and self.token: # as X-Esri-Auth header with generated token
                self._session.headers.update({token_header: "Bearer %s" % self.token})

        if try_json:
            params['f'] = 'json'

        resp = self._session.delete(url=url, data=params)
        return self._handle_response(resp=resp,
                                     out_path=out_path,
                                     file_name=file_name,
                                     try_json=try_json,
                                     force_bytes=kwargs.pop('force_bytes', False))
    #----------------------------------------------------------------------
    def streaming_method(self, url, callback,
                         data=None, json_data=None, verb="GET",
                         **kwargs):
        """
        Performs streaming web requests.

        =======================     ===========================================================
        **Parameters**              **Description**
        -----------------------     -----------------------------------------------------------
        url                         Required String. The web resource location.
        -----------------------     -----------------------------------------------------------
        callback                    Required Method.  The callback function to handle the response from the streaming request.

                                    **Example**
                                    ```
                                    def hook(r, *args, **kwargs):
                                        print('called a hook')
                                        return r
                                    ```
                                    See: https://requests.kennethreitz.org/en/master/user/advanced/#event-hooks


        -----------------------     -----------------------------------------------------------
        data                        Optional Dict. The parameters to pass to the method.
        -----------------------     -----------------------------------------------------------
        json_data                   Optional Dict. The parameters to pass to the method. This applies to POST only
        -----------------------     -----------------------------------------------------------
        verb                        Optional String.  The default is GET.  The allowed values are POST, GET, or PUT.
        -----------------------     -----------------------------------------------------------
        kwargs                      Optional Dict.  See https://requests.readthedocs.io/en/master/user/advanced/#request-and-response-objects
        =======================     ===========================================================

        """
        verbs = ['post', 'put', 'get']
        if verb.lower() in verbs:
            hooks = {'response' : callback}
            fn = getattr(self._session, verb.lower())
            if verb.lower() == "post":
                return fn(url=url,
                          data=data,
                          json_data=json,
                          hooks=hooks,
                          stream=True,
                          **kwargs)
            else:
                return fn(url=url,
                          data=data,
                          hooks=hooks,
                          stream=True,
                          **kwargs)
        else:
            allowed_verb = ",".join(verbs)
            raise ValueError(f"Invalid web method only {allowed_verb} as allowed")
        self._session.post(url=url, data=data, json_data=json, stream=True)
    #----------------------------------------------------------------------
    def login(self, username, password, expiration=None):
        """allows a user to login to a site with different credentials"""
        if expiration is None:
            expiration = 1440
        try:
            if self._username != username and \
               self._password != password:
                c = Connection(baseurl=self._baseurl,
                               username=username,
                               password=password)
                self = c
        except:
            raise Exception("Could not create a new login.")
    #----------------------------------------------------------------------
    def relogin(self, expiration=None):
        """ Re-authenticates with the portal using the same username/password. """
        if expiration is None:
            expiration = self._expiration
        self.logout()
        return self.token
    #----------------------------------------------------------------------
    def logout(self):
        """ Logs out of the portal. """
        self._token = None
        self._create_time = None
    #----------------------------------------------------------------------
    @property
    def is_logged_in(self):
        """ Returns true if logged into the portal. """
        return (self._auth in ['ANON', 'UNKNOWN']) == False
    #----------------------------------------------------------------------
    @property
    def product(self):
        """ Returns true if logged into the portal. """
        return self._product
    #----------------------------------------------------------------------
    @property
    def token(self):
        """Gets a Token"""
        if str(self._auth).lower() in ['builtin', 'oauth']:
            if self._expiration is None or self._expiration <= 5:
                self._expiration = 6
            if self._create_time and \
               (datetime.datetime.now() < self._create_time + datetime.timedelta(minutes=self._expiration)):
                return self._token
            elif str(self._auth).lower() == "oauth":
                self._create_time = datetime.datetime.now()
                self._token = self._oauth_token()
                return self._token
            elif self._create_time is None and \
               self._auth == "BUILTIN" and \
               self._product in ['SERVER', 'FEDERATED_SERVER']:
                self._token = self._server_token()
                return self._token
            elif self._product in ['FEDERATED_SERVER', 'GEOEVENT'] and \
                 self._portal_connection:
                self._token = self._server_token()
                return self._token
            elif self._create_time is None and \
                 self._auth == "OAUTH":
                self._token = self._oauth_token()
                return self._token
            elif self._create_time is None and \
                 self._product in ['PORTAL', "AGOL",
                                   "AGO", "ENTERPRISE"] and \
                 self._auth == "BUILTIN":
                self._token = self._enterprise_token()
                return self._token
            else:
                self._create_time = None
                return self.token
        elif str(self._auth.lower()) in ['home']:
            if self._create_time is None:
                self._create_time = datetime.datetime.now()
            if self._expiration is None:
                self._expiration = 1440
            if datetime.datetime.now() >= self._create_time + datetime.timedelta(minutes=self._expiration):
                raise Exception("Token is Expired. Please relogin to portal")
            return self._token
        elif str(self._auth).lower() == 'pro' or\
             self._baseurl.lower() == 'pro':
            self._create_time = datetime.datetime.now()
            self._token = self._pro_token()
            return self._token
        return None
    #----------------------------------------------------------------------
    @token.setter
    def token(self, value):
        """gets/sets the token"""
        if self._token != value:
            self._token = value
    #----------------------------------------------------------------------
    def _pro_token(self):
        """gets the token for various products"""
        if self._auth.lower() == 'pro' and \
           HASARCPY:
            resp = arcpy.GetSigninToken()
            if resp:
                if 'referer' in resp:
                    self._referer = resp['referer']
                if self._session:
                    self._session.headers['Referer'] = self._referer
                else:
                    self._referer = resp['referer']
                    self._session = self._create_session()
                if 'token' in resp:
                    return resp['token']
            else:
                raise Exception(("Could not login using Pro authencation."
                                 "Please verify in Pro that you are logged in."))
        return
    #----------------------------------------------------------------------
    def _server_token(self):
        """generates a server token"""
        if self._token_url is None and \
           self._portal_connection is None and \
           self._product in ['SERVER', "FEDERATED_SERVER"]:
            parsed = urlparse(self.baseurl)
            b = parsed.netloc
            if len(parsed.path) > 1:
                wa = parsed.path[1:].split('/')[0]
            else:
                wa = 'arcgis'
            ep = "admin/generateToken"
            self._token_url = "https://%s/%s/%s" % (b, wa, ep)
        elif self._token_url is None and \
             self._portal_connection:
            token_url = self._portal_connection._baseurl
            p = urlparse(token_url)
            self._token_url = "%s://%s/%s/sharing/rest/generateToken" % ("https",
                                                                         p.netloc,
                                                                         p.path[1:].split('/')[0],)
        if self._portal_connection:
            #self._token_url = token_url
            if self._portal_connection._auth.lower() == 'home':
                self._referer = ""
            ptoken = self._portal_connection.token
            postdata = {'serverURL':self._baseurl,
                        'token': ptoken,
                        'expiration': str(self._expiration),
                        'f': 'json',
                        'request':'getToken',
                        'referer':self._referer}
        else:
            postdata = { 'username': self._username,
                         'password': self._password,
                         #'client': 'requestip',
                         'referer' : self._referer,
                         'expiration': self._expiration,
                         'f': 'json' }
        res = self.post(path=self._token_url,
                        params=postdata,
                        ssl=True,
                        add_token=False)
        if 'error' in res:
            raise Exception(res['error'])
        self._create_time = datetime.datetime.fromtimestamp(int(res['expires'])/1000) - datetime.timedelta(minutes=self._expiration)
        self._token = res['token']
        return res['token']
    #----------------------------------------------------------------------
    def _enterprise_token(self):
        """generates a portal/agol token"""
        if self._referer is None and self._portal_connection is None:
            self._referer = "http"
        elif self._referer is None and self._portal_connection and \
             self._portal_connection._auth.lower() == "home":
            self._referer = ""
        postdata = { 'username': self._username, 'password': self._password,
                     'client': 'referer', 'referer': self._referer,
                     'expiration': self._expiration, 'f': 'json' }
        if self._token_url is None:
            if self._product in ['AGO', 'AGOL']:
                self._token_url = "https://%s/sharing/rest/generateToken" % urlparse(self._baseurl).netloc
            else:
                parsed = urlparse(self._baseurl)
                path = parsed.path
                if path.startswith('/'):
                    path = path[1:]
                self._token_url = "https://%s/%s/sharing/rest/generateToken" % (parsed.netloc,
                                                                                path.split('/')[0])

        res = self.post(path=self._token_url,
                        params=postdata,
                        add_token=False)
        if 'error' in res:
            raise Exception(res['error'])
        self._create_time = datetime.datetime.fromtimestamp(res['expires']/1000) - datetime.timedelta(minutes=self._expiration)
        return res['token']
    #----------------------------------------------------------------------
    def _oauth_token(self):
        """performs the oauth2 when secret and client exist"""
        from oauthlib.oauth2 import BackendApplicationClient
        from requests_oauthlib import OAuth2Session
        auth_url = '%soauth2/authorize' % self.baseurl
        tu = '%soauth2/token' % self.baseurl
        # handles the refreshing of the token
        if not (self._create_time is None) and \
           (datetime.datetime.now() >= self._create_time + datetime.timedelta(minutes=self._expiration)):
            self._token = None
        elif not (self._create_time is None) and \
             not (self._token is None) and \
             (datetime.datetime.now() < self._create_time + datetime.timedelta(minutes=self._expiration)):
            return self._token
        # Handles token generation
        if self._refresh_token is not None and \
           self._client_id is not None and\
           self._token is None: # Case 1: Refreshing a token
            parameters = {
                'client_id': self._client_id,
                'grant_type': 'refresh_token',
                'refresh_token': self._refresh_token,
                'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
            }
            token_info = self.post('oauth2/token', parameters, ssl=True, add_token=False)
            self._token = token_info['access_token']
            return self._token
        elif self._client_id and self._client_secret: # case 2: has both client and secret keys
            client = BackendApplicationClient(client_id=self._client_id)
            oauth = OAuth2Session(client=client)
            res = oauth.fetch_token(
                #method="GET",
                token_url=tu,
                client_id=self._client_id,
                client_secret=self._client_secret,
                include_client_id=True,
            )
            if 'expires_in' in res:
                self._create_time = datetime.datetime.fromtimestamp(res['expires_at']) - datetime.timedelta(seconds=7200)
                self._expiration = res['expires_in'] / 60
                if 'token' in res:
                    return res['token']
                if 'access_token' in res:
                    return res['access_token']
        elif self._client_id and \
             self._username is None and \
             self._password is None: # case 3: client id only

            auth_url = '%s/oauth2/authorize' % self.baseurl
            tu = '%s/oauth2/token' % self.baseurl
            oauth = OAuth2Session(self._client_id,
                                  redirect_uri="urn:ietf:wg:oauth:2.0:oob")
            authorization_url, state = oauth.authorization_url(auth_url)
            print("Please sign in to your GIS and paste the code that is obtained below.")
            print("If a web browser does not automatically open, please navigate to the URL below yourself instead.")
            print("Opening web browser to navigate to: " + authorization_url)
            import webbrowser, getpass
            webbrowser.open_new(authorization_url)
            authorization_response = getpass.getpass("Enter code obtained on signing in using SAML: ")

            self._create_time = datetime.datetime.now()
            token_info = oauth.fetch_token(tu,
                                           code=authorization_response,
                                           verify=False,
                                           include_client_id=True,
                                           authorization_response='authorization_code')
            self._expiration = token_info['expires_in']/60 - 2
            self._refresh_token = token_info['refresh_token']
            self._token = token_info['access_token']
            return self._token
        elif self._client_id and \
             not (self._username is None and \
                  self._password is None): # case 4: client id and username/password (SAML workflow)
            import re
            import json
            from bs4 import BeautifulSoup

            parameters = {
                'client_id': self._client_id,
                'response_type': 'code',
                'expiration': -1, # we want refresh_token to work for the life of the script
                'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
            }
            content = self.get('oauth2/authorize', parameters, ssl=True, try_json=False, add_token=False)

            pattern = re.compile('var oAuthInfo = ({.*?});', re.DOTALL)
            if len(pattern.findall(content)) == 0:
                pattern = re.compile('var oAuthInfo = ({.*?})', re.DOTALL)
            soup = BeautifulSoup(content, 'html.parser')
            for script in soup.find_all('script'):
                script_code = str(script.string).strip()
                matches = pattern.search(script_code)
                if not matches is None:
                    js_object = matches.groups()[0]
                    try:
                        oauth_info = json.loads(js_object)
                    except:
                        oauth_info = json.loads(js_object + "}")
                    break

            parameters = {
                'user_orgkey': '',
                'username': self._username,
                'password': self._password,
                'oauth_state': oauth_info['oauth_state']
            }
            content = self.post('oauth2/signin', parameters, ssl=True, try_json=False, add_token=False)
            soup = BeautifulSoup(content, 'html.parser')

            if soup.title is not None:
                if 'SUCCESS' in soup.title.string:
                    code = soup.title.string[len('SUCCESS code='):]

            oauth = OAuth2Session(self._client_id,
                                  redirect_uri="urn:ietf:wg:oauth:2.0:oob")
            if code is None:
                raise Exception("Could not generate a token.")
            self._create_time = datetime.datetime.now()
            token_info = oauth.fetch_token(tu, code=code, verify=False, include_client_id=True,
                                           authorization_response='authorization_code')
            self._refresh_token = token_info['refresh_token']
            self._token = token_info['access_token']
            self._expiration = token_info['expires_in']/60 - 2

            return self._token
        return None
    #----------------------------------------------------------------------
    def generate_portal_server_token(self, serverUrl, expiration=1440):
        """generates a server token using Portal token"""
        if self._auth.lower() in ["pki", 'iwa']:
            postdata = {'request': 'getToken',
                        'serverURL':serverUrl,
                        'referer': self._referer,
                        'f': 'json'}
            if expiration:
                postdata['expiration'] = expiration
        else:
            token = self.token
            postdata = {'serverURL':serverUrl,
                    'token': token,
                    'expiration':str(expiration),
                    'f': 'json',
                    'request':'getToken',
                    'referer':self._referer}
        if self._token_url is None:
            if self.baseurl.endswith('/'):
                resp = self.post('generateToken', postdata,
                                 ssl=True, add_token=False)
            else:
                resp = self.post('/generateToken', postdata,
                                 ssl=True, add_token=False)
        else:
            resp = self.post(path=self._token_url, postdata=postdata,
                             ssl=True, add_token=False)
        if isinstance(resp, dict) and resp:
            return resp.get('token')
        else:
            raise Exception(f"Could not generate the token for the service. \n Error Message: \n {resp}")
    #----------------------------------------------------------------------
    def _oauth_authenticate(self):
        """performs oauth check when only client_id is provided"""
        from urllib.parse import urlencode
        client_id = self._client_id
        expiration = self._expiration or 1440
        parameters = {
            'client_id': client_id,
            'response_type': 'code',
            'expiration': -1, # we want refresh_token to work for the life of the script
            'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
        }

        code = None

        if self._username is not None and self._password is not None: # built-in user through OAUTH
            content = self.get(self._token_url,
                               parameters,
                               ssl=True, try_json=False, add_token=False)
            import re
            import json
            from bs4 import BeautifulSoup
            pattern = re.compile('var oAuthInfo = ({.*?});', re.DOTALL)
            soup = BeautifulSoup(content, 'html.parser')
            for script in soup.find_all('script'):
                script_code = str(script.string).strip()
                matches = pattern.search(script_code)
                if not matches is None:
                    js_object = matches.groups()[0]
                    oauth_info = json.loads(js_object)
                    break

            parameters = {
                'user_orgkey': '',
                'username': self._username,
                'password': self._password,
                'oauth_state': oauth_info['oauth_state']
            }
            content = self.post('oauth2/signin', parameters, ssl=True, try_json=False, add_token=False)
            soup = BeautifulSoup(content, 'html.parser')

            if soup.title is not None:
                if 'SUCCESS' in soup.title.string:
                    code = soup.title.string[len('SUCCESS code='):]

        if code is None: # try interactive signin
            url = self._token_url#self._baseurl + 'oauth2/authorize'
            paramstring = urlencode(parameters)
            codeurl = "{}?{}".format(url, paramstring)

            import webbrowser
            import getpass

            print("Please sign in to your GIS and paste the code that is obtained below.")
            print("If a web browser does not automatically open, please navigate to the URL below yourself instead.")
            print("Opening web browser to navigate to: " + codeurl)
            webbrowser.open_new(codeurl)
            code = getpass.getpass("Enter code obtained on signing in using SAML: ")

        if code is not None:
            parameters = {
                'client_id': client_id,
                'grant_type': 'authorization_code',
                'code': code,
                'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
            }
            token_info = self.post(self._token_url, parameters, ssl=True, add_token=False)
            # print('******' + str(token_info))

            self._refresh_token = token_info['refresh_token']
            self._token = token_info['access_token']

            return self._token
        else:
            print("Unable to sign in using OAUTH")
            return None

    #----------------------------------------------------------------------
    def _check_product(self):
        """
        determines if the product is portal, arcgis online or arcgis server
        """
        from urllib.error import HTTPError
        baseurl = self._baseurl
        if self._product == "SERVER":
            if self._token_url is None:
                parsed = urlparse(self._baseurl)
                path = parsed.path[1:].split('/')[0]
                self._token_url = "https://%s/%s/admin/generateToken" % (parsed.netloc,
                                                                         path)

            return "SERVER"
        if baseurl is None:
            return "UNKNOWN"
        if baseurl.lower().find("arcgis.com") > -1:
            parsed = urlparse(self._baseurl)
            self._token_url = "https://%s/sharing/rest/generateToken" % parsed.netloc
            return "AGOL"
        elif baseurl.lower().find("/sharing/rest") > -1:
            if baseurl.endswith("/"):
                res = self.get(baseurl + 'info', params={'f' : 'json'}, add_token=False)
            else:
                res = self.get(baseurl + '/info', params={'f' : 'json'}, add_token=False)
            if self._token_url is None and \
               res is not None and \
               isinstance(res, dict) and \
               'authInfo' in res and \
               'tokenServicesUrl' in res['authInfo'] and \
               res['authInfo']['isTokenBasedSecurity']:
                parsed_from_system = urlparse(res['authInfo']['tokenServicesUrl'])
                parsed = urlparse(baseurl)
                if parsed.netloc.lower() != parsed_from_system.netloc.lower() and \
                   parsed.netloc.find(":7443") > -1: # WA not being used for token url
                    self._token_url = os.path.join(
                        parsed_from_system.scheme + "://",
                        parsed.netloc  + "/arcgis/" + "/".join(parsed_from_system.path[1:].split("/")[1:])
                    )
                    url_test = self._session.post(self._token_url, {'f': 'json'}, allow_redirects=False)
                    if url_test.status_code == 301:
                        self._token_url = url_test.headers['location']
                        if self._baseurl != os.path.dirname(url_test.headers['location']):
                            self._baseurl = os.path.dirname(url_test.headers['location'])
                else:
                    self._token_url = res['authInfo']['tokenServicesUrl']
            elif self._token_url is None and \
                 res is not None and \
                 isinstance(res, dict) and \
                 'authInfo' in res and \
                 'tokenServicesUrl' in res['authInfo'] and \
                 res['authInfo']['isTokenBasedSecurity'] == False:
                self._token_url = None
                self._auth = "OTHER"
            return "PORTAL"
        else:
            #Brute Force Method
            parsed = urlparse(baseurl)
            root = fr"{parsed.scheme}://{parsed.netloc}/{parsed.path[1:].split(r'/')[0]}"
            #root = baseurl.lower().split("/sharing")[0]
            #root = baseurl.lower().split('/rest')[0]
            parts = ['/info', '/rest/info', '/sharing/rest/info']
            params = {"f" : "json"}
            for pt in parts:
                try:
                    #print(pt)
                    res = self.get(root + pt, params=params, add_token=False)
                    if self._token_url is None and \
                       res is not None and \
                       isinstance(res, dict) and \
                       'authInfo' in res and \
                       'tokenServicesUrl' in res['authInfo'] and \
                       res['authInfo']['isTokenBasedSecurity']:
                        self._token_url = res['authInfo']['tokenServicesUrl']
                    elif self._token_url is None and \
                         res is not None and \
                         isinstance(res, dict) and \
                         'authInfo' in res and \
                         'tokenServicesUrl' in res['authInfo'] and \
                         res['authInfo']['isTokenBasedSecurity'] == False:
                        self._token_url = None
                        self._auth = "OTHER"
                except HTTPError as e:
                    res = ""
                except json.decoder.JSONDecodeError:
                    res = ""
                except Exception as e:
                    res = ""
                if isinstance(res, dict) and \
                   "currentVersion" in res and \
                   self._token_url:
                    t_parsed = urlparse(self._token_url[1:]).path
                    b_parsed = urlparse(self._baseurl[1:]).path
                    if t_parsed.startswith("/"):
                        t_parsed = t_parsed[1:].split("/")[0]
                    else:
                        t_parsed = t_parsed.split("/")[0]
                    if b_parsed.startswith("/"):
                        b_parsed = b_parsed[1:].split("/")[0]
                    else:
                        b_parsed = b_parsed.split("/")[0]
                    if t_parsed.lower() != b_parsed.lower():
                        self._token_url = None
                        if self._portal_connection:
                            return "FEDERATED_SERVER"
                        else:
                            from arcgis.gis import GIS
                            self._portal_connection = GIS(url=res['authInfo']['tokenServicesUrl'].split("/sharing/")[0],
                                                          username=self._username, password=self._password,
                                                          verify_cert=self._verify_cert)._con
                            return "FEDERATED_SERVER"
                    return "SERVER"
                elif isinstance(res, dict) and \
                     "currentVersion" in res and \
                     self._token_url is None:
                    return "SERVER"
                del pt
                del res
        return "PORTAL"





import logging
from .env import HAS_TENSORFLOW
if HAS_TENSORFLOW:
    import tensorflow as tf
    from tensorflow.keras.layers import Input, Conv2D, Dropout, ReLU, BatchNormalization, UpSampling2D, Reshape, Layer
    from tensorflow.keras import Model
    from .._utils.fastai_tf_fit import _tf_to_pytorch, _pytorch_to_tf_batch, _pytorch_to_tf
    from .common_tf import NormalizationLayerRGB, UpSample2DToSize

try:
    import torch
    import fastai
    import numpy as np
    from fastai.vision import imagenet_stats
    HAS_FASTAI = True
except:
    HAS_FASTAI = False


## Common start ##



## Common end ##

## Tensorflow specific utils start ##

def flatten_conv(output, n_anchors_per_cell, data_format):
    if data_format == 'channels_last':
        x = output
    else:
        x = tf.transpose(output, perm=[0, 2, 3, 1])
    x = Reshape((-1, x.shape[-1]//n_anchors_per_cell))(x)
    return x

def get_ssd_head_output(ssd):
    data_format = tf.keras.backend.image_data_format() #'channels_first' or 'channels_last'
    if data_format == 'channels_last':
        bn_axis = -1
    else:
        bn_axis = 1

    n_anchors_per_cell = ssd._anchors_per_cell

    ## SSD Head

    # Zero Layer
    head_block0_drop1 = Dropout(0.3, name="head_block0_drop1")(ssd._backbone_initalized.output)

    # First Layer
    head_block1_conv1 = Conv2D(256, 3, data_format=data_format, padding='same', activation='relu', name="head_block1_conv1")(head_block0_drop1)
    head_block1_bn1 = BatchNormalization(axis=bn_axis, name="head_block1_bn1")(head_block1_conv1)
    head_block1_drop1 = Dropout(0.3, name="head_block1_drop1")(head_block1_bn1)

    outputs = []

    # Create Output Pipeline for each grid_size
    for grid in ssd.grids:

        # Grid upsampling Block
        head_block2_conv1 = Conv2D(256, 3, data_format=data_format, padding='same', activation='relu', name=f"head_block2_conv1_{grid}")(head_block1_drop1)
        head_block2_bn1 = BatchNormalization(axis=bn_axis, name=f"head_block2_bn1_{grid}")(head_block2_conv1)
        head_block2_resize1 = UpSample2DToSize(grid, name=f"head_block2_resize1_{grid}")(head_block2_bn1)
        head_block2_drop1 = Dropout(0.3, name=f"head_block2_drop1_{grid}")(head_block2_resize1)

        # Final Activations layer
        conv1 = Conv2D(ssd._data.c*n_anchors_per_cell, 3, data_format=data_format, padding='same', name=f"head_block3_conv1_{grid}")     
        head_block3_conv1 = conv1(head_block2_drop1)
        conv2 = Conv2D(4*n_anchors_per_cell, 3, data_format=data_format, padding='same', name=f"head_block3_conv2_{grid}")
        head_block3_conv2 = conv2(head_block2_drop1)

        #conv1.bias = conv1.bias - 4.

        head_block3_output1 = flatten_conv(head_block3_conv1, n_anchors_per_cell, data_format)
        head_block3_output2 = flatten_conv(head_block3_conv2, n_anchors_per_cell, data_format)

        outputs.append((head_block3_output1, head_block3_output2))

    output1 = tf.concat(
        [a[0] for a in outputs],
        1,
        name='concat1'
    )
    output2 = tf.concat(
        [a[1] for a in outputs],
        1,
        name='concat2'
    )
    return output1, output2

def remove_padded_y(image_y_bboxes, image_y_classes):
    try:
        bboxes = tf.reshape(image_y_bboxes, (-1, 4))
    except Exception:
        bboxes = tf.zeros((0, 4))
    # bboxes_to_keep = ( ( bboxes[:,2] - bboxes[:,0] ) != 0 ) tf 2.0.0
    bboxes_to_keep = tf.not_equal(bboxes[:,2] - bboxes[:,0], 0)
    return image_y_bboxes[bboxes_to_keep],image_y_classes[bboxes_to_keep]

def activations_to_bboxes(activations, anchors, grid_sizes):
    activation_bboxes = tf.tanh(activations)
    activation_centers = (activation_bboxes[:,:2]/2 * grid_sizes) + anchors[:,:2]
    activation_height_widths = (activation_bboxes[:,2:]/2+1) * anchors[:,2:]
    return height_width_to_corner(activation_centers, activation_height_widths)

def height_width_to_corner(centers, height_width):
    x = tf.concat([centers - (height_width/2), centers + (height_width/2)], axis=1)
    return x

def one_hot_embedding(labels, num_classes):
    return tf.gather(tf.eye(num_classes), labels)

def tf_loss_function_single_image(ssd, image_y_bboxes, image_y_classes, image_p_bboxes, image_p_classes ):
    image_y_bboxes, image_y_classes = remove_padded_y(image_y_bboxes, image_y_classes)
    
    # denormalize_bboxes
    image_y_bboxes = (image_y_bboxes + 1.) / 2

    activation_bbox_corners = activations_to_bboxes(image_p_bboxes, ssd._anchors.cpu().numpy(), ssd._grid_sizes.cpu().numpy())
    
    overlaps = ssd._jaccard(_tf_to_pytorch(image_y_bboxes), ssd._anchor_cnr.data)
    try:
        gt_overlap, gt_idx = ssd._map_to_ground_truth(overlaps,False)
        gt_idx = gt_idx.numpy()
    except Exception as e:
        return tf.constant(0.), tf.constant(0.)
    gt_clas = tf.gather(image_y_classes, gt_idx)
    pos = (gt_overlap > 0.4)
    pos_idx = torch.nonzero(pos)[:,0].numpy()
    # gt_clas = tf.where( (pos.numpy()==0), 0, gt_clas) tf 2.0.0
    gt_clas = tf.compat.v2.where( (pos.numpy()==0), 0, gt_clas)
    ground_truth_classes = one_hot_embedding(gt_clas, ssd._data.c)
    gt_bbox = tf.gather(image_y_bboxes, gt_idx)

    location_loss = tf.reduce_sum(tf.abs( tf.gather(activation_bbox_corners, pos_idx) - tf.gather(gt_bbox, pos_idx) ))
    classification_loss = ssd._loss_function_classification(ground_truth_classes[:,1:], image_p_classes[:,1:])
    return location_loss, classification_loss

class postProcessingLayer(Layer):
    def __init__(self, grid_sizes, anchors):
        super(postProcessingLayer, self).__init__()
        self.grid_sizes = grid_sizes
        self.anchors = anchors
        
    def __call__(self, preds):
        class_value_store = []
        confidence_store = []
        bbox_store = []
        batch_size = preds[0].shape[0]
        if batch_size is None:
            batch_size = 1
        for i in range(batch_size):
            class_values, confidence_scores, bboxes = TFOD_post_process_predictions((preds[0][i], preds[1][i]), self.anchors, self.grid_sizes)
            class_value_store.append(class_values)
            confidence_store.append(confidence_scores)
            bbox_store.append(bboxes)
        return tf.stack(class_value_store), tf.stack(confidence_store), tf.stack(bbox_store)

def get_TFOD_post_processed_model(arcgis_model, input_normalization=True):
    model = arcgis_model.learn.model
    grid_sizes = _pytorch_to_tf(arcgis_model._grid_sizes)
    anchors = _pytorch_to_tf(arcgis_model._anchors)
    input_layer = model.input
    model_output = model.output

    if input_normalization:
        input_layer = Input(tuple(input_layer.shape[1:]))
        x = NormalizationLayerRGB()(input_layer)
        model_output = model(x)
    output_layer = postProcessingLayer(grid_sizes, anchors)(model_output)
    new_model = Model(input_layer, output_layer)
    return new_model

def TFOD_post_process_predictions(activations, anchors, grid_sizes):
    activation_classes, activation_bboxes = activations
    class_values = tf.argmax(activation_classes[:, 1:], axis=1)
    confidence_scores = tf.reduce_max(tf.sigmoid(activation_classes), axis=1)
    bboxes = TFOD_activations_to_bboxes(activation_bboxes, anchors, grid_sizes)
    return class_values, confidence_scores, bboxes

def TFOD_activations_to_bboxes(activation_bboxes, anchors, grid_sizes):
    activation_bboxes = tf.tanh(activation_bboxes)
    activation_centers = (activation_bboxes[:, :2]/2 * grid_sizes) + anchors[:, :2]
    activation_hw = (activation_bboxes[:, 2:]/2+1) * anchors[:, 2:]
    bboxes = tf.concat([activation_centers-activation_hw/2, activation_centers+activation_hw/2], axis=1)
    return bboxes

## Tensorflow specific utils end ##


try:
    import os, json
    import numpy as np
    import torch
    import math
    from torch import tensor
    HAS_TORCH = True

except Exception:
    HAS_TORCH = False

import arcgis
from arcgis.learn.models._multi_task_road_extractor import MultiTaskRoadExtractor

try:
    import arcpy
except Exception:
    pass


def calculate_rectangle_size_from_batch_size(batch_size):
    """
    calculate number of rows and cols to composite a rectangle given a batch size
    :param batch_size:
    :return: number of cols and number of rows
    """
    rectangle_height = int(math.sqrt(batch_size) + 0.5)
    rectangle_width = int(batch_size / rectangle_height)

    if rectangle_height * rectangle_width > batch_size:
        if rectangle_height >= rectangle_width:
            rectangle_height = rectangle_height - 1
        else:
            rectangle_width = rectangle_width - 1

    if (rectangle_height + 1) * rectangle_width <= batch_size:
        rectangle_height = rectangle_height + 1
    if (rectangle_width + 1) * rectangle_height <= batch_size:
        rectangle_width = rectangle_width + 1

    # swap col and row to make a horizontal rect
    if rectangle_height > rectangle_width:
        rectangle_height, rectangle_width = rectangle_width, rectangle_height

    if rectangle_height * rectangle_width != batch_size:
        return batch_size, 1

    return rectangle_height, rectangle_width


def get_tile_size(model_height, model_width, padding, batch_height, batch_width):
    """
    Calculate request tile size given model and batch dimensions
    :param model_height:
    :param model_width:
    :param padding:
    :param batch_width:
    :param batch_height:
    :return: tile height and tile width
    """
    tile_height = (model_height - 2 * padding) * batch_height
    tile_width = (model_width - 2 * padding) * batch_width

    return tile_height, tile_width


def tile_to_batch(
    pixel_block, model_height, model_width, padding, fixed_tile_size=True, **kwargs
):
    inner_width = model_width - 2 * padding
    inner_height = model_height - 2 * padding

    band_count, pb_height, pb_width = pixel_block.shape
    pixel_type = pixel_block.dtype

    if fixed_tile_size is True:
        batch_height = kwargs["batch_height"]
        batch_width = kwargs["batch_width"]
    else:
        batch_height = math.ceil((pb_height - 2 * padding) / inner_height)
        batch_width = math.ceil((pb_width - 2 * padding) / inner_width)

    batch = np.zeros(
        shape=(batch_width * batch_height, band_count, model_height, model_width),
        dtype=pixel_type,
    )
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        # pixel block might not be the shape (band_count, model_height, model_width)
        sub_pixel_block = pixel_block[
            :,
            y * inner_height: y * inner_height + model_height,
            x * inner_width: x * inner_width + model_width,
        ]
        sub_pixel_block_shape = sub_pixel_block.shape
        batch[
            b, :, : sub_pixel_block_shape[1], : sub_pixel_block_shape[2]
        ] = sub_pixel_block

    return batch, batch_height, batch_width


def batch_to_tile(batch, batch_height, batch_width):
    batch_size, bands, inner_height, inner_width = batch.shape
    tile = np.zeros(
        shape=(bands, inner_height * batch_height, inner_width * batch_width),
        dtype=batch.dtype,
    )

    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        tile[
            :,
            y * inner_height: (y + 1) * inner_height,
            x * inner_width: (x + 1) * inner_width,
        ] = batch[b]

    return tile


class ChildImageClassifier:
    def initialize(self, model, model_as_file):

        if not HAS_TORCH:
            raise Exception(
                "PyTorch is not installed. Install it using conda install -c pytorch pytorch torchvision"
            )

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info["ModelFile"]
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        self.road_extractor = MultiTaskRoadExtractor.from_model(
            data=None, emd_path=model
        )
        self.model = self.road_extractor.learn.model.to(self.device)
        self.model.eval()

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "padding",
                    "dataType": "numeric",
                    "value": int(self.json_info["ImageHeight"]) // 4,
                    "required": False,
                    "displayName": "Padding",
                    "description": "Padding",
                },
                {
                    "name": "batch_size",
                    "dataType": "numeric",
                    "required": False,
                    "value": 4,
                    "displayName": "Batch Size",
                    "description": "Batch Size",
                },
                {
                    "name": "predict_background",
                    "dataType": "string",
                    "required": False,
                    "value": "True",
                    "displayName": "Predict Background",
                    "description": "If False, will never predict the background/NoData Class.",
                },
            ]
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.padding = int(
            scalars.get("padding", self.json_info["ImageHeight"] // 4)
        )  # Default padding Imageheight//4.
        self.batch_size = (
            int(math.sqrt(int(scalars.get("batch_size", 4)))) ** 2
        )  # Default 4 batch_size
        self.predict_background = scalars.get("predict_background", "true").lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  # Default value True

        (
            self.rectangle_height,
            self.rectangle_width,
        ) = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            self.rectangle_height,
            self.rectangle_width,
        )

        return {
            "extractBands": tuple(self.json_info["ExtractBands"]),
            "padding": self.padding,
            "tx": tx,
            "ty": ty,
            "fixedTileSize": 1,
        }

    def updatePixels(self, tlc, shape, props, **pixelBlocks):  # 8 x 224 x 224 x 3
        input_image = pixelBlocks["raster_pixels"].astype(np.float32)
        batch, batch_height, batch_width = tile_to_batch(
            input_image,
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            fixed_tile_size=True,
            batch_height=self.rectangle_height,
            batch_width=self.rectangle_width,
        )

        semantic_predictions = self.pixel_classify_image(
            self.model,
            batch,
            self.device,
            classes=[clas["Name"] for clas in self.json_info["Classes"]],
            predict_bg=self.predict_background,
            model_info=self.json_info,
        )
        semantic_predictions = batch_to_tile(
            semantic_predictions.unsqueeze(dim=1).cpu().numpy(),
            batch_height,
            batch_width,
        )
        return semantic_predictions

    def pixel_classify_image(self, model, tiles, device, classes, predict_bg, model_info):
        model = model.to(device).eval()
        #logger.info("Tiles length is ", len(tiles))
        normed_batch_tensor = tensor(tiles).to(device).float()
        #logger.info("Normed length is ", len(normed_batch_tensor))
        with torch.no_grad():
            output, _ = model(normed_batch_tensor)
            #logger.debug("Output  is ", output)
            #logger.debug("Output  shape is ", output.shape)
        #_, road_pixels = torch.max(output, axis=1)
        #if not predict_bg:
        #    road_pixels[road_pixels == 0] = -1
        if predict_bg:
            return output.max(dim=1)[1]
        else:
            output[:, 0] = -1
            return output.max(dim=1)[1]
